import os
import typing
from pathlib import Path

import river.datasets.base

from cacp.comparison import DEFAULT_METRICS, DEFAULT_INCREMENTAL_METRICS
from cacp.comparison import process_comparison, process_incremental_comparison
from cacp.dataset import AVAILABLE_N_FOLDS, ClassificationDatasetBase, ClassificationFoldDataModifierBase
from cacp.info import dataset_info, classifier_info
from cacp.plot import process_comparison_results_plots, process_comparison_results_incremental_plots
from cacp.result import process_comparison_results
from cacp.time import process_times
from cacp.util import seed_everything
from cacp.wilcoxon import process_wilcoxon
from cacp.winner import process_comparison_result_winners


def run_experiment(
    datasets: typing.List[ClassificationDatasetBase],
    classifiers: typing.List[typing.Tuple[str, typing.Callable]],
    results_directory: typing.Union[str, os.PathLike] = './result',
    metrics: typing.Sequence[typing.Tuple[str, typing.Callable]] = DEFAULT_METRICS,
    n_folds: AVAILABLE_N_FOLDS = 10,
    custom_fold_modifiers: typing.List[ClassificationFoldDataModifierBase] = None,
    dob_scv: bool = True,
    categorical_to_numerical=True,
    normalized: bool = False,
    seed: int = 1,
    progress=lambda progress, total: None
):
    """
    [Main CACP Function] Runs automatic comparison of the performance evaluation of supervised classification
    algorithms by evaluating metrics on multiple datasets.

    :param datasets: dataset collection
    :param classifiers: classifiers collection
    :param results_directory: results directory
    :param metrics: metrics collection
    :param n_folds: number of folds {5,10}
    :param custom_fold_modifiers: custom fold modifiers that can change fold data before usage
    :param dob_scv: if folds distribution optimally balanced stratified cross-validation (DOB-SCV) should be used
    :param categorical_to_numerical: if dataset categorical values should be converted to numerical
    :param normalized: if the data should be normalized in range [0..1]
    :param seed: random seed value
    :param progress: function that can be used to monitor progress
    """
    seed_everything(seed)
    result_dir = Path(results_directory)
    result_dir.mkdir(exist_ok=True, parents=True)

    dataset_info(datasets, result_dir)
    classifier_info(classifiers, result_dir)
    process_comparison(
        datasets, classifiers, result_dir, metrics,
        n_folds=n_folds,
        dob_scv=dob_scv,
        categorical_to_numerical=categorical_to_numerical,
        normalized=normalized,
        custom_fold_modifiers=custom_fold_modifiers,
        progress=progress
    )
    process_comparison_results(result_dir, metrics)
    process_comparison_results_plots(result_dir, metrics)
    process_comparison_result_winners(result_dir, metrics)
    process_times(result_dir)
    process_wilcoxon(classifiers, result_dir, metrics)


def run_incremental_experiment(
    datasets: typing.List[typing.Union[ClassificationDatasetBase, river.datasets.base.Dataset]],
    classifiers: typing.List[typing.Tuple[str, typing.Callable]],
    results_directory: typing.Union[str, os.PathLike] = './result',
    metrics: typing.Sequence[typing.Tuple[str, typing.Callable]] = DEFAULT_INCREMENTAL_METRICS,
    seed: int = 1,
    progress=lambda progress, total: None,
):
    """
    [Main CACP Function] Runs automatic comparison of the performance evaluation of supervised classification
    algorithms by evaluating metrics on multiple datasets.

    :param datasets: dataset collection
    :param classifiers: classifiers collection
    :param results_directory: results directory
    :param metrics: metrics collection
    :param seed: random seed value
    :param progress: function that can be used to monitor progress

    """
    seed_everything(seed)
    result_dir = Path(results_directory)
    result_dir.mkdir(exist_ok=True, parents=True)

    dataset_info(datasets, result_dir)
    classifier_info(classifiers, result_dir)
    process_incremental_comparison(
        datasets, classifiers, result_dir, metrics, progress
    )

    process_comparison_results(result_dir, metrics)
    process_comparison_results_plots(result_dir, metrics)
    process_comparison_results_incremental_plots(result_dir, metrics)
    process_comparison_result_winners(result_dir, metrics)
    process_times(result_dir)
    process_wilcoxon(classifiers, result_dir, metrics)



import dataclasses
import typing
from abc import ABC, abstractmethod
from pathlib import Path
from urllib.request import urlretrieve
from zipfile import ZipFile

import numpy as np
import pandas as pd
import typing_extensions
from sklearn import preprocessing
from sklearn.model_selection import KFold
from tqdm import tqdm

BASE_KEEL_DATASETS_URL = 'https://github.com/sylwekczmil/cacp_files/raw/main/'

AVAILABLE_CLASSIFICATION_DATASET_NAMES = typing_extensions.Literal[
    'abalone', 'appendicitis', 'australian', 'automobile', 'balance', 'banana', 'bands',
    'breast', 'bupa', 'car', 'chess', 'cleveland', 'coil2000', 'contraceptive', 'crx',
    'dermatology', 'ecoli', 'flare', 'german', 'glass', 'haberman', 'hayes-roth', 'heart',
    'hepatitis', 'housevotes', 'ionosphere', 'iris', 'kr-vs-k', 'led7digit', 'letter',
    'lymphography', 'magic', 'mammographic', 'marketing', 'monk-2', 'movement_libras',
    'mushroom', 'newthyroid', 'nursery', 'optdigits', 'page-blocks', 'penbased', 'phoneme',
    'pima', 'post-operative', 'ring', 'saheart', 'satimage', 'segment', 'shuttle', 'sonar',
    'spambase', 'spectfheart', 'splice', 'tae', 'texture', 'thyroid', 'tic-tac-toe',
    'titanic', 'twonorm', 'vehicle', 'vowel', 'wdbc', 'wine', 'winequality-red',
    'winequality-white', 'wisconsin', 'yeast', 'zoo'
]

AVAILABLE_N_FOLDS = typing_extensions.Literal[5, 10]


@dataclasses.dataclass
class ClassificationFoldData:
    """
    Class that represents single dataset fold.
    """

    index: int = dataclasses.field()
    labels: np.ndarray = dataclasses.field()

    x_train: np.ndarray = dataclasses.field(repr=False)
    y_train: np.ndarray = dataclasses.field(repr=False)
    x_test: np.ndarray = dataclasses.field(repr=False)
    y_test: np.ndarray = dataclasses.field(repr=False)


class ClassificationFoldDataModifierBase(ABC):

    @abstractmethod
    def modify(self, fold: ClassificationFoldData) -> ClassificationFoldData:
        pass


class ClassificationFoldDataNormalizer(ClassificationFoldDataModifierBase):

    def modify(self, fold: ClassificationFoldData) -> ClassificationFoldData:
        x_tra_len = len(fold.x_train)
        x = np.concatenate([fold.x_train.astype(float), fold.x_test.astype(float)])
        min_max_scaler = preprocessing.MinMaxScaler()
        x = min_max_scaler.fit_transform(x)
        x_train, x_test = x[:x_tra_len], x[x_tra_len:]
        return ClassificationFoldData(
            index=fold.index,
            labels=fold.labels,
            x_train=x_train,
            y_train=fold.y_train,
            x_test=x_test,
            y_test=fold.y_test
        )


class ClassificationDatasetMinimalBase(ABC):
    """
    Minimal base class for classification dataset that represents single dataset.
    """

    def __init__(self, seed=1):
        self.seed = seed

    @abstractmethod
    def folds(
        self,
        n_folds: AVAILABLE_N_FOLDS = 10,
        dob_scv: bool = True,
        categorical_to_numerical=True
    ) -> typing.Iterable[ClassificationFoldData]:
        pass

    def __iter__(self):
        random_state = np.random.RandomState(seed=self.seed)
        for fold in self.folds():
            idx = random_state.permutation(np.arange(len(fold.x_test)))
            x_test = fold.x_test[idx]
            y_test = fold.y_test[idx]
            for x_data, y in zip(x_test, y_test):
                x = {i: value for i, value in enumerate(x_data)}
                yield x, y


class ClassificationDatasetBase(ClassificationDatasetMinimalBase):
    """
    Base class for classification dataset that represents single dataset.
    """

    @property
    @abstractmethod
    def name(self) -> str:
        pass

    @property
    @abstractmethod
    def instances(self) -> int:
        pass

    @property
    @abstractmethod
    def features(self) -> int:
        pass

    @property
    @abstractmethod
    def classes(self) -> int:
        pass

    def __str__(self):
        return f'Dataset name: {self.name}, ' \
               f'instances: {self.instances}, ' \
               f'features: {self.features}, ' \
               f'classes: {self.classes}'


class ClassificationDatasetDownloadProgressBar(tqdm):
    def update_to(self, b=1, bsize=1, t_size=None):
        if t_size is not None:
            self.total = t_size
        self.update(b * bsize - self.n)


class ClassificationDataset(ClassificationDatasetBase):
    """
    Class that represents KEEL single dataset.
    """

    def __init__(
        self, name: AVAILABLE_CLASSIFICATION_DATASET_NAMES,
        files_cache_path=Path.home().joinpath('cacp_files'),
        seed=1,
    ):
        """
        Initializes class instance that represents KEEL single dataset.

        :param name: KEEL dataset name
        :param files_cache_path: optional cache file patch where dataset will be downloaded
        """

        super().__init__(seed)

        self._name = name

        self._instances = 0
        self._features = 0
        self._classes = 0

        self._output = 'Class'
        self._origin = ''
        self._attributes: typing.Dict[str, str] = {}

        self._files_cache_path = files_cache_path
        self._files_cache_path.mkdir(exist_ok=True, parents=True)

        self._load_description()

    @property
    def name(self) -> str:
        return self._name

    @property
    def instances(self) -> int:
        return self._instances

    @property
    def features(self) -> int:
        return self._features

    @property
    def classes(self) -> int:
        return self._classes

    @property
    def origin(self) -> str:
        return self._origin

    @property
    def output_name(self) -> str:
        return self._output_name

    def folds(
        self,
        n_folds: AVAILABLE_N_FOLDS = 10,
        dob_scv: bool = True,
        categorical_to_numerical=True
    ) -> typing.Iterator[ClassificationFoldData]:

        zip_data_name = f'{self.name}-{n_folds}-{"dobscv" if dob_scv else "fold"}'
        data_path = self._fetch_data(zip_data_name, dob_scv)
        if dob_scv:
            data_name = f'{self.name}-{n_folds}dobscv'
        else:
            data_name = f'{self.name}-{n_folds}'

        for fold_index in range(1, n_folds + 1):
            train_data_path = data_path.joinpath(f'{data_name}-{fold_index}tra.dat')
            x_tra, y_tra = self._load_data(train_data_path, categorical_to_numerical)

            test_data_path = data_path.joinpath(f'{data_name}-{fold_index}tst.dat')
            x_tst, y_tst = self._load_data(test_data_path, categorical_to_numerical)

            labels = np.unique(np.hstack([y_tra, y_tst]))

            yield ClassificationFoldData(
                index=fold_index,
                x_train=x_tra,
                y_train=y_tra,
                x_test=x_tst,
                y_test=y_tst,
                labels=labels
            )

    def _load_description(self):
        file_name = f'{self.name}-names.txt'
        file_path = self._fetch_file(file_name)

        attributes_names = []
        attributes_types_names = []
        inputs = []
        output_name = 'Class'
        # KEEL descriptions files contain latin1 chars
        with file_path.open('r', encoding='latin1') as file:
            for line in file:
                if '@attribute' in line or '@Attribute' in line:
                    if '{' in line:
                        attr_name = line.split('{')[0].split()[1]
                        attr_type = 'category'
                    else:
                        s = line.split()[1:]
                        attr_name = s[0].strip()
                        attr_type = s[1].split('[')[0].strip()
                    attributes_names.append(attr_name)
                    attributes_types_names.append(attr_type)
                if '@input' in line:
                    inputs.append(line.split()[1:])
                elif '@output' in line:
                    output_name = line.split()[1]
                elif 'Origin.' in line:
                    self._origin = line.split('Origin.')[1].strip()
                elif 'Features.' in line:
                    self._features = int(line.split('Features.')[1].strip())
                elif 'Classes.' in line:
                    self._classes = int(line.split('Classes.')[1].strip())
                elif 'Instances.' in line:
                    self._instances = int(line.split('Instances.')[1].split()[0].strip())

        self._attributes = {n: t for n, t in zip(attributes_names, attributes_types_names)}
        self._output_name = output_name

    def _load_data(self, path: Path, categorical_to_numerical: bool) -> typing.Tuple[np.ndarray, np.ndarray]:
        skip_rows = 4 + len(self._attributes)
        df = pd.read_csv(path, skiprows=skip_rows, names=self._attributes.keys(), na_values='?')
        if categorical_to_numerical:
            for attr_name, attr_type_name in self._attributes.items():
                if attr_type_name == 'category':
                    df[attr_name] = df[attr_name].astype('category').cat.codes.values

        y = df[self._output_name].values
        del df[self._output_name]
        x = df.values
        return x, y

    def _fetch_data(self, data_name: str, dob_scv: bool) -> Path:
        data_path = self._files_cache_path.joinpath(data_name)
        data_unzip_path = data_path
        if dob_scv:
            data_path = data_path.joinpath(self.name)
            data_unzip_path = data_path.parent

        if not data_path.exists():
            zip_file_path = self._fetch_file(f'{data_name}.zip')
            with ZipFile(zip_file_path, mode='r') as zipfile:
                zipfile.extractall(data_unzip_path)
        return data_path

    def _fetch_file(self, file_name: str) -> Path:
        out_file_path = self._files_cache_path.joinpath(file_name)
        if not out_file_path.exists():
            url = f'{BASE_KEEL_DATASETS_URL}{file_name}'
            with ClassificationDatasetDownloadProgressBar(unit='B', unit_scale=True, miniters=1,
                                                          desc=f'Downloading {file_name}') as t:
                urlretrieve(url, filename=out_file_path, reporthook=t.update_to)

        return out_file_path


class LocalClassificationDataset(ClassificationDataset):
    """
    Class that represents single local dataset that has similar structure to KEEL dataset.
    """

    def __init__(self, name: str, dataset_directory: Path):
        """
        Initializes class instance that represents KEEL single local dataset.

        :param name: dataset name
        :param dataset_directory: directory where dataset is stored
        """
        super().__init__(name, dataset_directory)

    def _fetch_data(self, data_name: str, dob_scv: bool) -> Path:
        return self._files_cache_path

    def _fetch_file(self, file_name: str) -> Path:
        return self._files_cache_path.joinpath(file_name)


class LocalCsvClassificationDataset(ClassificationDatasetBase):
    """
    Class that represents single local dataset that is SCV with header.
    """

    @property
    def name(self) -> str:
        return self._name

    @property
    def instances(self) -> int:
        return self._instances

    @property
    def features(self) -> int:
        return self._features

    @property
    def classes(self) -> int:
        return self._classes

    def folds(self, n_folds: AVAILABLE_N_FOLDS = 10, dob_scv: bool = True, categorical_to_numerical=True) -> \
        typing.Iterable[ClassificationFoldData]:
        df = self._df()

        if categorical_to_numerical:
            for attr_name, attr_type_name in zip(df.columns, [t.name for t in df.dtypes]):
                if attr_type_name == 'category' or attr_type_name == 'object':
                    df[attr_name] = df[attr_name].astype('category').cat.codes.values

        y = df[self._output_name].values
        labels = np.unique(y)
        del df[self._output_name]
        x = df.values

        kf = KFold(n_splits=n_folds, shuffle=True, random_state=self.seed)
        for i, (train_index, test_index) in enumerate(kf.split(x), start=1):
            x_train, x_test = x[train_index], x[test_index]
            y_train, y_test = y[train_index], y[test_index]

            yield ClassificationFoldData(
                index=i,
                labels=labels,
                x_test=x_test,
                y_test=y_test,
                x_train=x_train,
                y_train=y_train
            )

    def _df(self):
        return pd.read_csv(self._dataset_path)

    def _load_metadata(self):
        df = self._df()
        self._instances = len(df)
        self._features = len(df.columns) - 1
        self._output_name = df.columns[-1]
        self._classes = len(df[self._output_name].unique())

    def __init__(self, name: str, dataset_path: Path):
        """
        Initializes class instance that represents CSV local dataset.

        :param name: dataset name
        :param dataset_path:path where dataset is stored
        """
        super().__init__()
        self._name = name
        self._output_name = ""
        self._dataset_path = dataset_path
        self._instances = 0
        self._features = 0
        self._classes = 0
        self._load_metadata()


def all_datasets() -> typing.List[ClassificationDataset]:
    """
    Gets all available datasets

    :return: all classification datasets
    """
    return [
        ClassificationDataset(name) for name in typing_extensions.get_args(AVAILABLE_CLASSIFICATION_DATASET_NAMES)
    ]



def process_unique_names(names: list):
    new_names = names.copy()
    for x in set(new_names):
        number = 0
        for i in range(0, len(new_names)):
            if new_names[i] == x:
                number += 1
                if number >= 2:
                    new_names[i] += str(number)
    return new_names



import pathlib
import typing
from pathlib import Path

import pandas as pd
import river.datasets.base

from cacp.dataset import ClassificationDatasetBase
from cacp.util import to_latex


def dataset_info(datasets: typing.Iterable[typing.Union[ClassificationDatasetBase, river.datasets.base.Dataset]],
                 result_dir: Path):
    """
    Produces results files with list of all datasets used in experiment alog with their attributes.

    :param datasets: dataset collection
    :param result_dir: results directory

    """
    records = []
    for dataset_idx, dataset in enumerate(datasets):
        name = dataset.__class__.__name__.lower()
        if hasattr(dataset, "name"):
            name = dataset.name

        i = 0
        x = {}
        labels = set()
        for i, (x, y) in enumerate(dataset):
            labels.add(y)

        row = {
            'Dataset': name,
            'Instances': i + 1,
            'Features': len(x),
            'Classes': len(labels),
        }
        records.append(row)

    df = pd.DataFrame(records)
    df.index += 1
    info_dir = result_dir.joinpath('info')
    info_dir.mkdir(exist_ok=True, parents=True)
    df.to_csv(info_dir.joinpath('datasets.csv'), index=True)
    f = info_dir.joinpath('datasets.tex').open('w')
    tex = to_latex(
        df,
        caption='Datasets used to perform experiments',
        label='tab:datasets'
    )
    f.write(tex)


def classifier_info(classifiers: typing.Iterable[typing.Tuple[str, typing.Callable]], result_dir: pathlib.Path):
    """
    Produces results files with list of all classifiers used in experiment along with their attributes.

    :param classifiers: classifiers collection
    :param result_dir: results directory

    """
    records = []
    for cn, c in classifiers:
        cz = c(2, 2).__class__
        library = cz.__module__.split('.')[0]
        name = cz.__name__
        classifier_type = 'CUSTOM'
        if library == 'sklearn':
            classifier_type = 'BATCH'
        elif library == 'skmultiflow':
            classifier_type = 'INCREMENTAL'
        elif library == 'river':
            classifier_type = 'INCREMENTAL'

        row = {
            'Name': cn,
            'Class Name': name,
            'Library': library,
            'Type': classifier_type
        }
        records.append(row)

    df = pd.DataFrame(records)
    df = df.drop_duplicates()
    df.sort_values(['Type', 'Name'], inplace=True)
    df = df.reset_index(drop=True)
    df.index += 1
    info_dir = result_dir.joinpath('info')
    info_dir.mkdir(exist_ok=True, parents=True)
    df.to_csv(info_dir.joinpath('classifiers.csv'), index=True)
    f = info_dir.joinpath('classifiers.tex').open('w')
    tex = to_latex(
        df,
        caption='Classifiers used to perform experiments',
        label='tab:algorithm'
    )
    f.write(tex)



from abc import abstractmethod
from functools import lru_cache
from pydoc import locate
from typing import Type, Dict
from typing import TypeVar, Generic, List

from pydantic import BaseModel

from cacp.gui.external.shared.schema import parse_model
from cacp.gui.external.shared.type import get_all_non_abstract_subclasses, to_id

T = TypeVar("T")

BASE_DATE_FORMAT = "%d/%m/%Y %H:%M:%S"


class BaseAppModel(BaseModel, Generic[T]):

    @classmethod
    @abstractmethod
    def all(cls) -> List[T]:
        return []

    @classmethod
    @abstractmethod
    def get_by_id(cls, _id: str) -> T:
        return None


class ClassModel(BaseAppModel, Generic[T]):
    id: str

    @classmethod
    @abstractmethod
    def from_class(cls, source_class: Type) -> T:
        raise NotImplementedError()

    @classmethod
    @abstractmethod
    def base_class(cls):
        raise NotImplementedError()

    @property
    def class_type(self) -> Type:
        return locate(self.id)

    @property
    def json_schema(self) -> dict:
        try:
            model: Type[BaseModel] = parse_model(self.class_type)
            schema = model.schema()
            return schema
        except Exception as e:
            print(e)

    @classmethod
    @lru_cache
    def all(cls) -> List[T]:
        return list(cls.all_dict().values())

    @classmethod
    @lru_cache
    def all_dict(cls) -> Dict[str, T]:
        result = {}
        for sub_class in get_all_non_abstract_subclasses(cls.base_class()):
            try:
                model: T = cls.from_class(sub_class)
                model.test()
                _id = to_id(sub_class)
                model.id = _id
                result[_id] = model
            except Exception:
                # do not add those that can not be initialized
                pass
        return result

    @classmethod
    @lru_cache
    def get_by_id(cls, _id: str) -> T:
        return cls.all_dict().get(_id)

    def test(self):
        defaults = {}
        schema = self.json_schema
        for k, v in schema.get("properties").items():
            if default := v.get("default"):
                defaults[k] = default
        self.class_type(**defaults)



import dash
import dash_ag_grid as dag
import dash_bootstrap_components as dbc
from dash import html, callback, Output, Input, State, no_update

from cacp.gui.components.shared.utils import global_location_href_output
from cacp.gui.db.experiments import get_all_experiments, delete_experiment
from cacp.gui.preview import preview_button_kwargs

dash.register_page(__name__, path="/")

EXPERIMENTS_TABLE_ID = "experiments_table"


def layout():
    init_experiments = get_all_experiments()

    return html.Div([
        html.Div(
            [dbc.Button("Create new experiment", href="/experiment-form")],
            className="d-flex justify-content-end mb-4"),
        html.H5("Experiments"),
        dag.AgGrid(
            id=EXPERIMENTS_TABLE_ID,
            rowData=init_experiments,
            columnDefs=[
                {"field": "name"},
                {"field": "type"},
                {"field": "path"},
                {"field": "number of datasets"},
                {"field": "number of classifiers"},
                {"field": "status"},
                {"field": "created at"},
                {
                    "field": "remove", "headerName": "", "sortable": False, "filter": False, "resizable": False,
                    "cellRenderer": "Button",
                    "cellRendererParams": {
                        "buttonName": "Delete",
                        "buttonClassName": "btn btn-sm btn-danger",
                        "iconClassName": "bi bi-trash delete-button",
                        **preview_button_kwargs()
                    }
                },
                {
                    "field": "navigate", "headerName": "", "sortable": False, "filter": False, "resizable": False,
                    "cellRenderer": "Button",
                    "cellRendererParams": {
                        "buttonName": "Open",
                        "buttonClassName": "btn btn-sm btn-primary",
                        "iconClassName": "bi bi-forward-fill"
                    },
                },
            ],
            defaultColDef={"sortable": True, "filter": True, "resizable": True},
            columnSize="responsiveSizeToFit",
            style={"height": "80vh"}
        ),
    ])


@callback(
    Output(EXPERIMENTS_TABLE_ID, "rowData"),
    global_location_href_output(),
    Input(EXPERIMENTS_TABLE_ID, "cellRendererData"),
    State(EXPERIMENTS_TABLE_ID, "rowData"),
    prevent_initial_call=True,
)
def on_experiments_renderer_data(
    experiments_renderer_data,
    experiments
):
    if not experiments_renderer_data:
        return no_update, no_update

    idx = experiments_renderer_data["rowIndex"]
    experiment_id = experiments[idx]["id"]

    if experiments_renderer_data["colId"] == "navigate":
        return no_update, f"/experiment/{experiment_id}"

    if experiments_renderer_data["colId"] == "remove":
        delete_experiment(experiment_id)

    return get_all_experiments(), no_update



from __future__ import annotations

from functools import lru_cache
from inspect import signature
from typing import Type, get_type_hints

from pydantic import BaseModel, BaseConfig
from pydantic.fields import ModelField


def _new_pseudo_pydantic_model(name: str):
    class Model(BaseModel):
        pass

    Model.__name__ = name
    return Model


@lru_cache
def _try_parse(field_name: str, field_type_hint: Type, default_value):
    model = _new_pseudo_pydantic_model("test")
    try:
        field = ModelField.infer(
            name=field_name,
            value=default_value,
            annotation=field_type_hint,
            class_validators=None,
            config=BaseConfig,
        )
        model.__fields__[field_name] = field
        model.schema()
        return field
    except Exception:
        return None


def parse_model(t: Type) -> Type[bool | int | float | complex | str | BaseModel, None]:
    model = _new_pseudo_pydantic_model(t.__name__)
    init_signature = signature(t.__init__)
    init_type_hints = get_type_hints(t.__init__)
    init_parameters = init_signature.parameters
    for field_name, field_type_hint in init_type_hints.items():
        field_parameter = init_parameters[field_name]
        default_value = field_parameter.default if field_parameter.default != field_parameter.empty else None
        if field := _try_parse(field_name, field_type_hint, default_value):
            model.__fields__[field_name] = field
    return model



import contextlib
from functools import lru_cache
from typing import Type, List

import river
from river.metrics.base import ClassificationMetric

from cacp.gui.external.shared.model import ClassModel
from cacp.gui.external.shared.type import to_id


class RiverMetricModel(ClassModel):
    name: str
    docs_url: str

    @classmethod
    def base_class(cls):
        return ClassificationMetric

    @classmethod
    @lru_cache
    def all(cls) -> List["RiverMetricModel"]:
        result = []
        for k, v in cls.all_dict().items():
            with contextlib.suppress(NotImplementedError):
                m = v.class_type()
                if hasattr(m, "get") and isinstance(m.get(), float):
                    result.append(v)

        return result

    @classmethod
    def from_class(cls, source_class: Type) -> "RiverMetricModel":
        _id = to_id(source_class)
        docs_split = _id.split(".")
        docs_version = river.__version__
        docs_name = docs_split[3] if len(docs_split) > 3 else docs_split[2]
        return cls(
            id=_id,
            name=source_class.__name__,
            docs_url=f"https://riverml.xyz/{docs_version}/api/{docs_split[1].replace('_', '-')}/{docs_name}/"
        )



from pydoc import locate
from typing import Dict, Callable, cast

from cacp.gui.external.river_library.classifier import RiverClassifierModel
from cacp.gui.external.sklearn_library.classifier import SklearnClassifierModel


def map_classifiers(classifiers: list):
    return [
        {"#": i + 1, "docs_url": f"[{c.docs_url}]({c.docs_url})", "json_schema": c.json_schema, **c.dict()} for
        i, c in
        enumerate(sorted(classifiers, key=lambda c: c.name))
    ]


SKLEARN_CLASSIFIERS = map_classifiers(SklearnClassifierModel.all())
RIVER_CLASSIFIERS = map_classifiers(RiverClassifierModel.all())


def parse_classifier(classifier_dict: Dict):
    if "code" in classifier_dict:  # custom classifier
        classifier_type = cast(Callable, locate(classifier_dict["locate_id"], True))
        return lambda n_inputs, n_classes: classifier_type()
    else:
        classifier_type = cast(Callable, locate(classifier_dict["id"]))
        return lambda n_inputs, n_classes: classifier_type(**classifier_dict["init_values"])



import traceback

import dash_bootstrap_components as dbc
from dash import html, Output, callback, Input, no_update

from cacp import run_experiment, run_incremental_experiment
from cacp.gui.components.shared.utils import GLOBAL_LOCATION_ID
from cacp.gui.db.experiments import get_experiment, ExperimentStatus, ExperimentType, update_experiment_status
from cacp.gui.external.classifier import parse_classifier
from cacp.gui.external.dataset import parse_dataset
from cacp.gui.external.metric import parse_metric
from cacp.gui.external.shared.helpers import process_unique_names


def make_progress_graph(progress: int, total: int):
    if total == 0:
        return dbc.Progress(label="Loading", value=100, className="mt-4 mb-4", animated=True, striped=True)

    value = int(progress / (total + 1) * 100)
    return html.Div([
        html.P("Experiment is running, do not close this page", className="lead"),
        dbc.Progress(label=f"{value}%", value=value, className="mt-4 mb-4", animated=True, striped=True),
    ])


class ExperimentProcessor(html.Div):

    def __init__(
        self,
        experiment_store_id
    ):

        self.experiment_processor_progress_id = "experiment_processor_progress"
        super().__init__([
            html.Div(make_progress_graph(0, 0), id=self.experiment_processor_progress_id)
        ])

        @callback(
            Output(experiment_store_id, "data"),
            Input(GLOBAL_LOCATION_ID, "pathname"),
            prevent_initial_call=False,
            background=True,
            progress=Output(self.experiment_processor_progress_id, "children"),
            cancel=[Input(GLOBAL_LOCATION_ID, "href")],
        )
        def on_page_load(set_progress, pathname: str):
            result = no_update
            path_split = pathname.split("/")
            if len(path_split) > 2:
                experiment_id = int(path_split[2])
                experiment = get_experiment(experiment_id)
                if experiment and experiment["status"] == ExperimentStatus.RUNNING:
                    try:

                        def progress(p: int, t: int):
                            set_progress(make_progress_graph(p, t))

                        classifier_names = process_unique_names(
                            [c["name"] for c in experiment["classifiers"]]
                        )
                        classifiers = list(
                            zip(classifier_names, [parse_classifier(c) for c in experiment["classifiers"]]))
                        datasets = [parse_dataset(d) for d in experiment["datasets"]]
                        metric_names = process_unique_names(
                            [m["name"] for m in experiment["metrics"]]
                        )
                        metrics = list(zip(metric_names, [parse_metric(m) for m in experiment["metrics"]]))
                        experiment_type = experiment["type"]
                        if experiment_type == ExperimentType.BATCH:
                            update_experiment_status(experiment_id, ExperimentStatus.RUNNING)
                            run_experiment(
                                datasets,
                                classifiers,
                                experiment["path"],
                                metrics,
                                progress=progress,
                            )
                        elif experiment_type == ExperimentType.INCREMENTAL:
                            update_experiment_status(experiment_id, ExperimentStatus.RUNNING)
                            run_incremental_experiment(
                                datasets,
                                classifiers,
                                experiment["path"],
                                metrics,
                                progress=progress,
                            )
                        update_experiment_status(experiment_id, ExperimentStatus.FINISHED)
                    except Exception as e:
                        print(e, traceback.format_exc())
                        update_experiment_status(experiment_id, ExperimentStatus.FAILED)

                    result = get_experiment(experiment_id)

            set_progress("")
            return result



from pathlib import Path

import dash
import dash_ace
import dash_bootstrap_components as dbc
from dash import html, Output, callback, Input, State, no_update
from dash.dcc import Store

from cacp.gui.components.shared.utils import global_location_href_output
from cacp.gui.db.custom_datasets import get_custom_dataset, CustomDatasetType, CustomDataset, \
    update_custom_dataset, CUSTOM_DATASET_CODE_TEMPLATE, \
    test_custom_dataset_code, KEEL_DATASET_CODE_TEMPLATE, CSV_DATASET_CODE_TEMPLATE
from cacp.gui.preview import preview_button_kwargs

dash.register_page(__name__, path_template="/custom_dataset/<custom_dataset_id>")

CUSTOM_DATASET_NAME_ID = "custom_dataset_name_id"
CUSTOM_DATASET_TYPE_ID = "custom_dataset_type_id"
CUSTOM_DATASET_PATH_ID = "custom_dataset_path_id"
CUSTOM_DATASET_PATH_DIV_ID = "custom_dataset_path_div_id"
CUSTOM_DATASET_CODE_ID = "custom_dataset_code_id"
CUSTOM_DATASET_CODE_DIV_ID = "custom_dataset_code_div_id"
CUSTOM_DATASET_SAVE_BUTTON_ID = "custom_dataset_save_button_id"
CUSTOM_DATASET_TOAST_ID = "custom_dataset_toast_id"
CUSTOM_DATASET_STORE_ID = "custom_dataset_store_id"
CODE_HEIGHT = {"height": "calc(100vh - 450px)", "width": "100%"}
CODE_HEIGHT_INVISIBLE = {"height": "0", "width": "100%", "visibility": "hidden"}


def layout(custom_dataset_id=None):
    custom_dataset = get_custom_dataset(custom_dataset_id) if custom_dataset_id else {}

    name_input = html.Div(
        [
            dbc.Label("Name", html_for=CUSTOM_DATASET_NAME_ID),
            dbc.Input(id=CUSTOM_DATASET_NAME_ID, placeholder="Enter name", value=custom_dataset.get("name")),
            dbc.FormText(
                "Provide name for your dataset, it should be unique.",
                color="secondary",
            ),
        ],
        className="mb-3",
    )

    type_input = html.Div(
        [
            dbc.Label("Type", html_for=CUSTOM_DATASET_TYPE_ID),
            dbc.RadioItems(
                {
                    CustomDatasetType.CUSTOM_CODE: "Custom code",
                    CustomDatasetType.CSV_FILE: "CSV file (last column will be treated as class column)",
                    CustomDatasetType.KEEL_FILES: "Keel files (should have KEEL datasets structure)"
                },
                value=custom_dataset.get("type"),
                id=CUSTOM_DATASET_TYPE_ID
            )
        ],
        className="py-2",
    )

    path_input = html.Div(
        [
            dbc.Label("Path", html_for=CUSTOM_DATASET_PATH_ID),
            dbc.Input(id=CUSTOM_DATASET_PATH_ID, placeholder="Enter path", value=custom_dataset.get("path")),
        ],
        id=CUSTOM_DATASET_PATH_DIV_ID,
        className="d-none",
    )

    code = html.Div(
        [
            dbc.Label("Code", html_for=CUSTOM_DATASET_NAME_ID), dash_ace.DashAceEditor(
            id=CUSTOM_DATASET_CODE_ID,
            value=custom_dataset.get("code"),
            theme='github',
            mode='python',
            tabSize=2,
            enableBasicAutocompletion=True,
            placeholder='Python code ...',
            style={"height": "calc(100% - 30px)", "width": "100%"}
        )],
        id=CUSTOM_DATASET_CODE_DIV_ID,
        style=CODE_HEIGHT
    )

    save_button = html.Div([
        dbc.Button(
            "Save", id=CUSTOM_DATASET_SAVE_BUTTON_ID, className="mt-3", n_clicks=0, **preview_button_kwargs()
        )
    ], className="d-flex justify-content-end align-items-center")

    toast = dbc.Toast(
        "",
        id=CUSTOM_DATASET_TOAST_ID,
        header="Validation failed",
        is_open=False,
        dismissable=True,
        duration=5000,
        style={"position": "fixed", "top": 66, "right": 10, "width": 350, "z-index": "2000"},
    )

    store = Store(id=CUSTOM_DATASET_STORE_ID, data=custom_dataset)

    view = html.Div([
        name_input,
        type_input,
        path_input,
        code,
        save_button,
        toast,
        store
    ])

    return view


@callback(
    Output(CUSTOM_DATASET_PATH_DIV_ID, "className"),
    Input(CUSTOM_DATASET_TYPE_ID, "value"),
)
def update_path_display(type_value: CustomDatasetType):
    return "d-none" if type_value == CustomDatasetType.CUSTOM_CODE else "py-4"


@callback(
    Output(CUSTOM_DATASET_CODE_ID, "value"),
    Output(CUSTOM_DATASET_CODE_DIV_ID, "style"),
    Input(CUSTOM_DATASET_TYPE_ID, "value"),
    Input(CUSTOM_DATASET_NAME_ID, "value"),
    Input(CUSTOM_DATASET_PATH_ID, "value"),
    State(CUSTOM_DATASET_STORE_ID, "data")
)
def update_river_datasets_button_class(type_value: CustomDatasetType, name: str, path: str, dataset: CustomDataset):
    if dash.ctx.triggered_id == CUSTOM_DATASET_TYPE_ID:
        if type_value == CustomDatasetType.CUSTOM_CODE:
            return CUSTOM_DATASET_CODE_TEMPLATE.format(dataset.get("id")), CODE_HEIGHT
        if type_value == CustomDatasetType.CSV_FILE:
            return CSV_DATASET_CODE_TEMPLATE.format(dataset.get("id"), name, path), CODE_HEIGHT_INVISIBLE
        if type_value == CustomDatasetType.KEEL_FILES:
            return KEEL_DATASET_CODE_TEMPLATE.format(dataset.get("id"), name, path), CODE_HEIGHT_INVISIBLE

    if type_value == CustomDatasetType.CSV_FILE:
        return CSV_DATASET_CODE_TEMPLATE.format(dataset.get("id"), name, path), CODE_HEIGHT_INVISIBLE
    if type_value == CustomDatasetType.KEEL_FILES:
        return KEEL_DATASET_CODE_TEMPLATE.format(dataset.get("id"), name, path), CODE_HEIGHT_INVISIBLE
    return no_update, no_update


@callback(
    global_location_href_output(),
    Output(CUSTOM_DATASET_TOAST_ID, "children"),
    Output(CUSTOM_DATASET_TOAST_ID, "header"),
    Output(CUSTOM_DATASET_TOAST_ID, "is_open"),
    Input(CUSTOM_DATASET_SAVE_BUTTON_ID, "n_clicks"),
    State(CUSTOM_DATASET_NAME_ID, "value"),
    State(CUSTOM_DATASET_TYPE_ID, "value"),
    State(CUSTOM_DATASET_CODE_ID, "value"),
    State(CUSTOM_DATASET_PATH_ID, "value"),
    State(CUSTOM_DATASET_STORE_ID, "data"),
    prevent_initial_call=True,
)
def save_button_click(
    n_clicks, name_value: str, type_value: CustomDatasetType, code_value: str, path_value: str,
    custom_dataset: CustomDataset
):
    href, toast_message, toast_header, toast_is_open, = no_update, "", no_update, False
    if n_clicks:
        custom_dataset_id = int(custom_dataset["id"])
        errors = []
        if not name_value:
            errors.append("Name can not be empty.")

        if type_value != CustomDatasetType.CUSTOM_CODE:
            path = Path(path_value)
            if type_value == CustomDatasetType.CSV_FILE and not path.exists():
                errors.append("Path is invalid, should be existing file.")
            if type_value == CustomDatasetType.KEEL_FILES and (not path.exists() or not path.is_dir()):
                errors.append("Path is invalid for KEEL dataset, should be existing directory.")

        if not errors:
            test_error = test_custom_dataset_code(custom_dataset_id, name_value, code_value)
            if test_error:
                errors.append(test_error)

        if errors:
            toast_message = dbc.Alert([html.Div(e) for e in errors], color="danger")
            toast_header = "Validation error"
        else:
            update_custom_dataset(custom_dataset_id, name_value, type_value, code_value, path_value)
            toast_message = dbc.Alert(html.Div("Saved"))
            toast_header = "Success"
        toast_is_open = True
    return href, toast_message, toast_header, toast_is_open



import os


def is_preview():
    return os.environ.get("PREVIEW", "false") == "true"


def preview_prevent_modifications():
    if is_preview():
        raise RuntimeError("CACP is in PREVIEW mode")


def preview_button_kwargs():
    if is_preview():
        return dict(disabled=True)
    return {}



from functools import lru_cache
from typing import Type, List

import river
from river.base import Classifier

from cacp.gui.external.shared.model import ClassModel
from cacp.gui.external.shared.type import to_id


class RiverClassifierModel(ClassModel):
    name: str
    docs_url: str

    @classmethod
    def base_class(cls):
        return Classifier

    @classmethod
    @lru_cache
    def all(cls) -> List["RiverClassifierModel"]:
        return [c for c in cls.all_dict().values() if "Base" not in c.name]

    @classmethod
    def from_class(cls, source_class: Type) -> "RiverClassifierModel":
        _id = to_id(source_class)
        docs_split = _id.split(".")
        docs_version = river.__version__
        docs_name = docs_split[3] if len(docs_split) > 3 else docs_split[2]
        return cls(
            id=_id,
            name=source_class.__name__,
            docs_url=f"https://riverml.xyz/{docs_version}/api/{docs_split[1].replace('_', '-')}/{docs_name}/"
        )



import dash
from dash import html, callback, Input, no_update

from cacp.gui.components.classifiers.custom_classifiers_table import CustomClassifiersTable
from cacp.gui.components.classifiers.river_classifiers_table import RiverClassifiersTable
from cacp.gui.components.classifiers.sklearn_classifiers_table import SklearnClassifiersTable
from cacp.gui.components.shared.utils import global_location_href_output
from cacp.gui.db.custom_classifiers import add_custom_classifier
from cacp.gui.preview import preview_button_kwargs

dash.register_page(__name__)

ADD_CUSTOM_CLASSIFIER_BUTTON_ID = "add_custom_classifier_button"

layout = html.Div([
    html.Div(
        html.Button("Add custom classifier", className="btn btn-primary", id=ADD_CUSTOM_CLASSIFIER_BUTTON_ID,
                    **preview_button_kwargs()),
        className="d-flex justify-content-end mb-4"),
    html.H5("Custom classifiers"),
    CustomClassifiersTable("cct"),
    html.H5("Sklearn classifiers (BATCH)", className="mt-4"),
    SklearnClassifiersTable("sct"),
    html.H5("River classifiers (INCREMENTAL)", className="mt-4"),
    RiverClassifiersTable("rct"),
])


@callback(
    global_location_href_output(),
    Input(ADD_CUSTOM_CLASSIFIER_BUTTON_ID, "n_clicks"),
    prevent_initial_call=True,
)
def add_custom_classifier_button(n_clicks):
    if n_clicks:
        custom_classifier_id = add_custom_classifier()
        return f"/custom_classifier/{custom_classifier_id}"
    return no_update



from enum import Enum
from typing import Optional, Type

import river
from river.datasets.base import Dataset, SyntheticDataset

from cacp.gui.external.shared.model import ClassModel
from cacp.gui.external.shared.type import to_id


class RiverDataSetType(str, Enum):
    SYNTHETIC = "Synthetic"
    REAL = "Real"

    @classmethod
    def from_river_dataset(cls, river_dataset: Dataset):
        if issubclass(river_dataset.__class__, SyntheticDataset):
            return cls.SYNTHETIC
        else:
            return cls.REAL

    def docs_url(self):
        docs_version = river.__version__
        docs_api_sub_path = "synth" if self == RiverDataSetType.SYNTHETIC else "datasets"
        return f"https://riverml.xyz/{docs_version}/api/{docs_api_sub_path}/"


class RiverDatasetModel(ClassModel):
    name: str
    type: RiverDataSetType
    task: str
    samples: int
    features: int
    outputs: Optional[int] = None
    classes: Optional[int] = None
    docs_url: str

    @classmethod
    def base_class(cls):
        return Dataset

    @classmethod
    def from_class(cls, source_class: Type) -> "RiverDatasetModel":
        river_dataset: Dataset = source_class()
        dataset_repr = river_dataset._repr_content

        def parse_int_field(value):
            return int(dataset_repr.get(value, "0").replace("∞", "0").replace(",", ""))

        dataset_name = dataset_repr.get("Name")
        dataset_type = RiverDataSetType.from_river_dataset(river_dataset)

        task = dataset_repr.get("Task")
        classes = parse_int_field("Classes")
        if classes == 0 and task == "Binary classification":
            classes = 2

        return cls(
            id=to_id(source_class),
            name=dataset_name,
            type=dataset_type,
            task=dataset_repr.get("Task"),
            samples=parse_int_field("Samples"),
            features=parse_int_field("Features"),
            outputs=parse_int_field("Outputs"),
            classes=classes,
            docs_url=dataset_type.docs_url() + dataset_name
        )



from typing import List, Dict

import dash_bootstrap_components as dbc
from dash import html, Output, callback, Input, State, ctx, no_update

from cacp.gui.components.classifiers.custom_classifiers_table import CustomClassifiersTable
from cacp.gui.components.classifiers.river_classifiers_table import RiverClassifiersTable
from cacp.gui.components.classifiers.sklearn_classifiers_table import SklearnClassifiersTable
from cacp.gui.components.datasets.custom_datasets_table import CustomDatasetsTable
from cacp.gui.components.datasets.keel_datasets_table import KeelDatasetsTable
from cacp.gui.components.datasets.river_datasets_table import RiverDatasetsTable
from cacp.gui.components.experiments.new.selected_classifiers_table import SelectedClassifiersTable
from cacp.gui.components.experiments.new.selected_datasets_table import SelectedDatasetsTable
from cacp.gui.components.experiments.new.selected_metrics_table import SelectedMetricsTable
from cacp.gui.components.experiments.new.selection_modal import SelectionModal
from cacp.gui.components.metrics.custom_metrics_table import CustomMetricsTable
from cacp.gui.components.metrics.river_metrics_table import RiverMetricsTable
from cacp.gui.components.metrics.sklearn_metrics_table import SklearnMetricsTable
from cacp.gui.components.shared.utils import global_location_href_output
from cacp.gui.db.experiments import ExperimentType, add_experiment
from cacp.gui.preview import preview_button_kwargs


class NewExperimentForm(html.Div):
    class ids:
        table = lambda aio_id: f"NewExperimentForm-table-{aio_id}"
        name_input = lambda aio_id: f"NewExperimentForm-name_input-{aio_id}"
        type_input = lambda aio_id: f"NewExperimentForm-type_input-{aio_id}"
        selected_datasets_store = lambda aio_id: f"NewExperimentForm-selected_datasets_store-{aio_id}"
        selected_classifiers_store = lambda aio_id: f"NewExperimentForm-selected_classifiers_store-{aio_id}"
        selected_metrics_store = lambda aio_id: f"NewExperimentForm-selected_metrics_store-{aio_id}"
        toast = lambda aio_id: f"NewExperimentForm-toast-{aio_id}"
        create_button = lambda aio_id: f"NewExperimentForm-create_button-{aio_id}"

    ids = ids

    def __init__(
        self,
        aio_id
    ):

        name_input = html.Div(
            [
                dbc.Label("Name", html_for=self.ids.name_input(aio_id)),
                dbc.Input(id=self.ids.name_input(aio_id), placeholder="Enter name"),
                dbc.FormText(
                    "Provide name for your experiment, it should be unique.",
                    color="secondary",
                ),
            ],
            className="mb-3",
        )

        type_input = html.Div(
            [
                dbc.Label("Type", html_for=self.ids.type_input(aio_id)),
                dbc.RadioItems(
                    {
                        ExperimentType.BATCH: "Batch learning (k-fold Cross-Validation)",
                        ExperimentType.INCREMENTAL: "Incremental learning (Prequential evaluation)"
                    },
                    value=ExperimentType.BATCH,
                    id=self.ids.type_input(aio_id)
                ),
                dbc.FormText(
                    "The choice of the type of experiment determines which classifiers and datasets can be used later.",
                    color="secondary",
                ),
            ],
            className="py-2",
        )
        custom_datasets_selection = SelectionModal(
            "Add Custom dataset", CustomDatasetsTable, aio_id=f"{aio_id}-kd"
        )
        keel_datasets_selection = SelectionModal(
            "Add Keel dataset", KeelDatasetsTable, aio_id=f"{aio_id}-cd", button_kwargs=dict(className="mx-2"),
        )
        river_datasets_selection = SelectionModal(
            "Add River dataset", RiverDatasetsTable,
            button_kwargs=dict(className="d-none"), aio_id=f"{aio_id}-rd"
        )
        custom_classifiers_selection = SelectionModal(
            "Add Custom classifiers", CustomClassifiersTable, aio_id=f"{aio_id}-cc"
        )
        river_classifiers_selection = SelectionModal(
            "Add River classifiers", RiverClassifiersTable,
            button_kwargs=dict(className="d-none"), aio_id=f"{aio_id}-rc"
        )
        sklearn_classifiers_selection = SelectionModal(
            "Add Sklearn classifier", SklearnClassifiersTable,
            button_kwargs=dict(className="d-none"), aio_id=f"{aio_id}-sc"
        )
        custom_metrics_selection = SelectionModal(
            "Add Custom metric", CustomMetricsTable, aio_id=f"{aio_id}-cm"
        )
        river_metrics_selection = SelectionModal(
            "Add River metric", RiverMetricsTable,
            button_kwargs=dict(className="d-none"), aio_id=f"{aio_id}-rm"
        )
        sklearn_metrics_selection = SelectionModal(
            "Add Sklearn metric", SklearnMetricsTable,
            button_kwargs=dict(className="d-none"), aio_id=f"{aio_id}-sm"
        )

        selected_datasets_table = SelectedDatasetsTable(
            store_id=self.ids.selected_datasets_store(aio_id),
            aio_id=aio_id
        )
        selected_classifiers_table = SelectedClassifiersTable(
            store_id=self.ids.selected_classifiers_store(aio_id),
            aio_id=aio_id
        )
        selected_metrics_table = SelectedMetricsTable(
            store_id=self.ids.selected_metrics_store(aio_id),
            aio_id=aio_id
        )

        super().__init__([
            html.Div([
                name_input,
                type_input,
                html.Br(),
                html.Div([
                    custom_datasets_selection, keel_datasets_selection, river_datasets_selection,
                ]),
                html.Br(),
                html.H5("Selected datasets"),
                selected_datasets_table,
                html.Br(),
                html.Div([
                    custom_classifiers_selection, sklearn_classifiers_selection, river_classifiers_selection,
                ]),
                html.Br(),
                html.H5("Selected classifiers"),
                selected_classifiers_table,
                html.Br(),
                html.Div([
                    custom_metrics_selection, sklearn_metrics_selection, river_metrics_selection,
                ]),
                html.Br(),
                html.H5("Selected metrics"),
                selected_metrics_table,
                html.Div([
                    dbc.Button(
                        "Create", id=self.ids.create_button(aio_id), className="mt-3", n_clicks=0,
                        **preview_button_kwargs()
                    )
                ], className="d-flex justify-content-end align-items-center"),
                dbc.Toast(
                    "",
                    id=self.ids.toast(aio_id),
                    header="Validation failed",
                    is_open=False,
                    dismissable=True,
                    duration=5000,
                    icon="danger",
                    style={"position": "fixed", "top": 66, "right": 10, "width": 350, "z-index": "2000"},
                ),
            ])
        ])

        @callback(
            Output(sklearn_classifiers_selection.ids.modal_open_button(sklearn_classifiers_selection.aio_id),
                   "className"),
            Input(self.ids.type_input(aio_id), "value"),
        )
        def update_sklearn_classifiers_button_class(type_value):
            return "mx-2" if type_value == ExperimentType.BATCH else "d-none"

        @callback(
            Output(sklearn_metrics_selection.ids.modal_open_button(sklearn_metrics_selection.aio_id),
                   "className"),
            Input(self.ids.type_input(aio_id), "value"),
        )
        def update_sklearn_metricss_button_class(type_value):
            return "mx-2" if type_value == ExperimentType.BATCH else "d-none"

        @callback(
            Output(river_classifiers_selection.ids.modal_open_button(river_classifiers_selection.aio_id), "className"),
            Input(self.ids.type_input(aio_id), "value"),
        )
        def update_river_classifiers_button_class(type_value):
            return "mx-2" if type_value == ExperimentType.INCREMENTAL else "d-none"

        @callback(
            Output(river_datasets_selection.ids.modal_open_button(river_datasets_selection.aio_id), "className"),
            Input(self.ids.type_input(aio_id), "value"),
        )
        def update_river_datasets_button_class(type_value):
            return "mx-2" if type_value == ExperimentType.INCREMENTAL else "d-none"

        @callback(
            Output(river_metrics_selection.ids.modal_open_button(river_metrics_selection.aio_id), "className"),
            Input(self.ids.type_input(aio_id), "value"),
        )
        def update_river_metrics_button_class(type_value):
            return "mx-2" if type_value == ExperimentType.INCREMENTAL else "d-none"

        @callback(
            Output(self.ids.selected_classifiers_store(aio_id), "data"),
            Input(custom_classifiers_selection.store_id, "data"),
            Input(sklearn_classifiers_selection.store_id, "data"),
            Input(river_classifiers_selection.store_id, "data"),
            Input(selected_classifiers_table.ids.table(selected_classifiers_table.aio_id), "cellRendererData"),
            Input(self.ids.type_input(aio_id), "value"),
            State(self.ids.selected_classifiers_store(aio_id), "data")
        )
        def selected_classifier(
            selected_custom_classifier_data: list,
            selected_sklearn_classifier_data: list,
            selected_river_classifier_data: list,
            cell_renderer_data: dict,
            _type_value: ExperimentType,
            prev_data: list
        ):
            if ctx.triggered_id == self.ids.type_input(aio_id):
                return []
            elif ctx.triggered_id == custom_classifiers_selection.store_id and selected_custom_classifier_data:
                return prev_data + [cc for cc in selected_custom_classifier_data if cc["type"] == _type_value]
            elif ctx.triggered_id == sklearn_classifiers_selection.store_id and selected_sklearn_classifier_data:
                return prev_data + selected_sklearn_classifier_data
            elif ctx.triggered_id == river_classifiers_selection.store_id and selected_river_classifier_data:
                return prev_data + selected_river_classifier_data
            elif ctx.triggered_id == selected_classifiers_table.ids.table(selected_classifiers_table.aio_id):
                del prev_data[cell_renderer_data["rowIndex"]]
                return prev_data
            return no_update

        @callback(
            Output(self.ids.selected_datasets_store(aio_id), "data"),
            Input(custom_datasets_selection.store_id, "data"),
            Input(keel_datasets_selection.store_id, "data"),
            Input(river_datasets_selection.store_id, "data"),
            Input(selected_datasets_table.ids.table(selected_datasets_table.aio_id), "cellRendererData"),
            Input(self.ids.type_input(aio_id), "value"),
            State(self.ids.selected_datasets_store(aio_id), "data")
        )
        def selected_dataset(
            selected_custom_dataset_data: list,
            selected_keel_dataset_data: list,
            selected_river_dataset_data: list,
            cell_renderer_data: dict,
            _type_value: ExperimentType,
            prev_data: list
        ):

            if ctx.triggered_id == self.ids.type_input(aio_id):
                return []
            elif ctx.triggered_id == custom_datasets_selection.store_id and selected_custom_dataset_data:
                return prev_data + selected_custom_dataset_data
            elif ctx.triggered_id == keel_datasets_selection.store_id and selected_keel_dataset_data:
                return prev_data + [{**d, "name": d["Name"]} for d in selected_keel_dataset_data]
            elif ctx.triggered_id == river_datasets_selection.store_id and selected_river_dataset_data:
                return prev_data + selected_river_dataset_data
            elif ctx.triggered_id == selected_datasets_table.ids.table(selected_datasets_table.aio_id):
                del prev_data[cell_renderer_data["rowIndex"]]
                return prev_data
            return no_update

        @callback(
            Output(self.ids.selected_metrics_store(aio_id), "data"),
            Input(custom_metrics_selection.store_id, "data"),
            Input(sklearn_metrics_selection.store_id, "data"),
            Input(river_metrics_selection.store_id, "data"),
            Input(selected_metrics_table.ids.table(selected_metrics_table.aio_id), "cellRendererData"),
            Input(self.ids.type_input(aio_id), "value"),
            State(self.ids.selected_metrics_store(aio_id), "data")
        )
        def selected_metric(
            selected_custom_metric_data: list,
            selected_sklearn_metric_data: list,
            selected_river_metric_data: list,
            cell_renderer_data: dict,
            _type_value: ExperimentType,
            prev_data: list
        ):

            if ctx.triggered_id == self.ids.type_input(aio_id):
                return []
            elif ctx.triggered_id == custom_metrics_selection.store_id and selected_custom_metric_data:
                return prev_data + selected_custom_metric_data
            elif ctx.triggered_id == sklearn_metrics_selection.store_id and selected_sklearn_metric_data:
                return prev_data + selected_sklearn_metric_data
            elif ctx.triggered_id == river_metrics_selection.store_id and selected_river_metric_data:
                return prev_data + selected_river_metric_data
            elif ctx.triggered_id == selected_metrics_table.ids.table(selected_metrics_table.aio_id):
                del prev_data[cell_renderer_data["rowIndex"]]
                return prev_data
            return no_update

        @callback(
            global_location_href_output(),
            Output(self.ids.toast(aio_id), "children"),
            Output(self.ids.toast(aio_id), "is_open"),
            Input(self.ids.create_button(aio_id), "n_clicks"),
            State(self.ids.name_input(aio_id), "value"),
            State(self.ids.type_input(aio_id), "value"),
            State(self.ids.selected_datasets_store(aio_id), "data"),
            State(self.ids.selected_classifiers_store(aio_id), "data"),
            State(self.ids.selected_metrics_store(aio_id), "data"),
            prevent_initial_call=True,
        )
        def run_experiment_button(
            n_clicks, name_value: str, type_value: ExperimentType, selected_datasets: List[Dict],
            selected_classifiers: List[Dict], selected_metrics: List[Dict]
        ):
            href, toast_message, toast_is_open, = no_update, "", False
            if n_clicks:
                errors = []
                if not name_value:
                    errors.append("Name can not be empty.")
                if len(selected_datasets) == 0:
                    errors.append("At least 1 dataset need to be selected.")
                if len(selected_classifiers) < 2:
                    errors.append("At least 2 classifiers need to be selected.")
                if len(selected_metrics) == 0:
                    errors.append("At least 1 metric need to be selected.")
                if errors:
                    toast_message = dbc.Alert([html.Div(e) for e in errors], color="danger")
                    toast_is_open = True
                else:
                    experiment_id = add_experiment(name_value, type_value, selected_datasets, selected_classifiers,
                                                   selected_metrics)
                    href = f"/experiment/{experiment_id}"
            return href, toast_message, toast_is_open



import dash_bootstrap_components as dbc
from dash import html


def sidebar_component():
    return html.Div(
        [
            html.Div([
                html.Div([
                    html.H2(dbc.NavLink("CACP", href="/"), className="display-4"),
                    html.Hr(),
                    html.P(
                        "Classification Algorithms Comparison Pipeline", className="lead"
                    ),
                    dbc.Nav(
                        [
                            dbc.NavLink("Datasets", href="/datasets", active="exact"),
                            dbc.NavLink("Classifiers", href="/classifiers", active="exact"),
                            dbc.NavLink("Metrics", href="/metrics", active="exact"),
                            dbc.NavLink("Experiments", href="/", active="exact"),
                            html.Br(),
                        ],
                        vertical=True,
                        pills=True,
                    ),
                ]),
                dbc.Button("Create new experiment", href="/experiment-form"),
                html.Script("runNavLinksRoutine()"),
            ],
                className="d-flex justify-content-between flex-column h-100"),
        ],
        style={
            "position": "fixed",
            "top": 0,
            "left": 0,
            "bottom": 0,
            "width": "16rem",
            "padding": "2rem 1rem",
            "backgroundColor": "#f8f9fa",
        }
    )



import tempfile
import traceback
import typing
from datetime import datetime
from enum import Enum
from pathlib import Path
from typing import TypedDict, List

from river import metrics as river_metrics
from river.dummy import NoChangeClassifier
from sklearn.dummy import DummyClassifier
from tinydb import TinyDB
from tinydb.table import Document

from cacp import ClassificationDataset
from cacp.comparison import process_comparison_single, process_incremental_comparison_single
from cacp.gui.custom.metrics import CUSTOM_METRICS_CODE_DIR
from cacp.gui.external.metric import parse_metric
from cacp.gui.preview import preview_prevent_modifications
from cacp.util import accuracy


class CustomMetricType(str, Enum):
    BATCH = "BATCH"
    INCREMENTAL = "INCREMENTAL"


class CustomMetric(TypedDict):
    id: int
    locate_id: str
    name: str
    type: CustomMetricType
    code: str
    created_at: float


CUSTOM_METRICS_DB = TinyDB(CUSTOM_METRICS_CODE_DIR / "metrics.json")
CUSTOM_METRIC_BATCH_CODE_TEMPLATE = """import numpy as np


def metric{}(y_true: np.ndarray, y_pred: np.ndarray, labels: np.ndarray):  # do not change function declaration
    return 0.5
"""

CUSTOM_INCREMENTAL_METRIC_CODE_TEMPLATE = """from river.metrics.base import ClassificationMetric


class Metric{}(ClassificationMetric):  # do not change class declaration
    def get(self) -> float:
        return 0.5
"""


def _convert_from_document_to_custom_metric(custom_metric: Document) -> CustomMetric:
    if custom_metric:
        custom_metric["id"] = custom_metric.doc_id
        custom_metric["created at"] = datetime.fromtimestamp(custom_metric["created_at"]).strftime(
            "%Y-%m-%d %H:%M:%S")
        custom_metric["json_schema"] = {
            "title": custom_metric["name"],
            "type": "object",
            "properties": {
            }
        }
    return custom_metric


def _locate_id(custom_metric_id: int, type_value: CustomMetricType,
               original_custom_metric_id: typing.Optional[int] = None):
    if original_custom_metric_id is None:
        original_custom_metric_id = custom_metric_id

    if type_value == CustomMetricType.BATCH:
        return f"cacp.gui.custom.metrics.metric{custom_metric_id}.metric{original_custom_metric_id}"
    else:
        return f"cacp.gui.custom.metrics.metric{custom_metric_id}.Metric{original_custom_metric_id}"


def _code_path(custom_metric_id: int) -> Path:
    return CUSTOM_METRICS_CODE_DIR.joinpath(f"metric{custom_metric_id}.py")


def _save_code(custom_metric_id: int, code: str):
    with _code_path(custom_metric_id).open("w") as f:
        f.write(code)


def _remove_code(custom_metric_id: int):
    _code_path(custom_metric_id).unlink(missing_ok=True)


def get_all_custom_metrics() -> List[CustomMetric]:
    return [_convert_from_document_to_custom_metric(e) for e in CUSTOM_METRICS_DB.all() if e is not None]


def get_custom_metric(custom_metric_id: int) -> CustomMetric:
    return _convert_from_document_to_custom_metric(CUSTOM_METRICS_DB.get(doc_id=custom_metric_id))


def add_custom_metric() -> int:
    preview_prevent_modifications()
    new_custom_metric: CustomMetric = dict()
    all_metrics = CUSTOM_METRICS_DB.all()
    new_custom_metric_id = 1 if len(all_metrics) == 0 else all_metrics[-1].doc_id + 1
    new_custom_metric["name"] = f"Custom Metric {new_custom_metric_id}"
    new_custom_metric["type"] = CustomMetricType.BATCH
    new_custom_metric["code"] = CUSTOM_METRIC_BATCH_CODE_TEMPLATE.format(new_custom_metric_id)
    new_custom_metric["locate_id"] = _locate_id(new_custom_metric_id, CustomMetricType.BATCH)
    new_custom_metric["created_at"] = datetime.now().timestamp()
    _save_code(new_custom_metric_id, new_custom_metric["code"])
    return CUSTOM_METRICS_DB.insert(new_custom_metric)


def update_custom_metric(
    custom_metric_id: int,
    name_value: str,
    type_value: CustomMetricType,
    code_value: str
):
    preview_prevent_modifications()
    CUSTOM_METRICS_DB.update({"name": name_value, "type": type_value, "code": code_value,
                              "locate_id": _locate_id(custom_metric_id, type_value)},
                             doc_ids=[custom_metric_id])
    _save_code(custom_metric_id, code_value)


def test_custom_metric_code(custom_metric_id: int, code_value: str, type_value: CustomMetricType):
    preview_prevent_modifications()
    error = None
    _save_code(0, code_value)
    metric_factory = parse_metric({"locate_id": _locate_id(0, type_value, custom_metric_id), "code": code_value})

    try:
        ds = ClassificationDataset("iris")
        ds_fold = next(ds.folds())
        if type_value == CustomMetricType.BATCH:
            process_comparison_single(lambda n_inputs, n_classes: DummyClassifier(), "test", ds, ds_fold,
                                      (('Accuracy', accuracy), ("Custom", metric_factory)))
        elif type_value == CustomMetricType.INCREMENTAL:
            with tempfile.TemporaryDirectory() as tmpdirname:
                process_incremental_comparison_single(
                    lambda n_inputs, n_classes: NoChangeClassifier(), "test",
                    ds, 3, Path(tmpdirname),
                    (('Accuracy', river_metrics.Accuracy), ("Custom", metric_factory))
                )
    except:
        error = traceback.format_exc()

    _remove_code(0)
    return error


def delete_custom_metric(custom_metric_id: int):
    preview_prevent_modifications()
    CUSTOM_METRICS_DB.remove(doc_ids=[custom_metric_id])
    _remove_code(custom_metric_id)



import typing
from pathlib import Path

import pandas as pd

from cacp.comparison import DEFAULT_METRICS
from cacp.util import to_latex


def process_comparison_result_winners_for_metric(metric: str, result_dir: Path) -> pd.DataFrame:
    """
    Processes comparison results, finds winners for metric.

    :param metric: comparison metric {auc, accuracy, precision, recall, f1}
    :param result_dir: results directory
    :return: DateFrame with winners for metric

    """
    df = pd.read_csv(result_dir.joinpath('comparison.csv'))
    algorithms = df['Algorithm'].unique()
    places = [i for i in range(min(len(algorithms), 3))]

    winner_dir = result_dir.joinpath('winner').joinpath(metric.lower())
    winner_dir.mkdir(exist_ok=True, parents=True)

    def count_places(place=0):
        count = {a: 0 for a in algorithms}
        names = {a: [] for a in algorithms}
        for dataset, df_d in df.groupby('Dataset'):
            df_d_a_m = df_d.groupby('Algorithm').mean(numeric_only=True).sort_values(by=[metric], ascending=False)
            best = df_d_a_m.iloc[place]
            count[best.name] += 1
            names[best.name].append(dataset)

        return count, names

    counts = []
    for c, n in [count_places(i) for i in places]:
        counts.append(c)

    rows = []
    for algorithm in algorithms:
        row = [algorithm]
        for p in places:
            row.append(counts[p][algorithm])
        rows.append(row)

    columns = ['Algorithm'] + ['1st', '2nd', '3rd'][: len(places)]
    df_r = pd.DataFrame(columns=columns, data=rows)
    df_r = df_r.sort_values(by=['1st'], ascending=False)
    df_r.reset_index(drop=True, inplace=True)
    df_r.index += 1
    df_r.to_csv(winner_dir.joinpath('comparison_result.csv'), index=True)
    winner_dir.joinpath('comparison_result.tex').open('w').write(
        to_latex(
            df_r,
            caption=f'Ranking of compared algorithms for {metric}',
            label=f'tab:places_{metric}',
        )
    )

    return df_r


def process_comparison_result_winners(result_dir: Path,
                                      metrics: typing.Sequence[typing.Tuple[str, typing.Callable]] = DEFAULT_METRICS):
    """
    Processes comparison results, finds winners.

    :param result_dir: results directory
    :param metrics: metrics collection

    """
    wins_df = None
    for metric, _ in metrics:
        metric_wins = process_comparison_result_winners_for_metric(metric, result_dir).sort_values(by=['Algorithm'])
        if wins_df is None:
            wins_df = metric_wins[['Algorithm']].copy()

        for c in metric_wins.columns[1:]:
            wins_df[f'{metric} {c}'] = metric_wins[c].values

    winner_dir = result_dir.joinpath('winner')
    winner_dir.mkdir(exist_ok=True, parents=True)
    wins_df = wins_df.sort_values(by=wins_df.columns[1:].values.tolist(), ascending=False)
    wins_df.reset_index(drop=True, inplace=True)
    wins_df.index += 1
    wins_df.to_csv(winner_dir.joinpath('comparison.csv'), index=True)
    winner_dir.joinpath('comparison.tex').open('w').write(
        to_latex(
            wins_df,
            caption='Ranking of compared algorithms',
            label='tab:places',
        )
    )



import tempfile
import traceback
import typing
from datetime import datetime
from enum import Enum
from pathlib import Path
from typing import TypedDict, List

from river.dummy import NoChangeClassifier
from sklearn.dummy import DummyClassifier
from tinydb import TinyDB
from tinydb.table import Document

from cacp.comparison import process_comparison_single, DEFAULT_METRICS, process_incremental_comparison_single, \
    DEFAULT_INCREMENTAL_METRICS
from cacp.gui.custom.datasets import CUSTOM_DATASETS_CODE_DIR
from cacp.gui.external.dataset import parse_dataset
from cacp.gui.preview import preview_prevent_modifications


class CustomDatasetType(str, Enum):
    CUSTOM_CODE = "CUSTOM_CODE"
    CSV_FILE = "CSV_FILE"
    KEEL_FILES = "KEEL_FILES"


class CustomDataset(TypedDict):
    id: int
    locate_id: str
    name: str
    type: CustomDatasetType
    code: str
    path: typing.Optional[str]
    created_at: float


CUSTOM_DATASETS_DB = TinyDB(CUSTOM_DATASETS_CODE_DIR / "datasets.json")
CUSTOM_DATASET_CODE_TEMPLATE = """import typing

import numpy as np
from sklearn.model_selection import KFold

from cacp import ClassificationDatasetMinimalBase, ClassificationFoldData
from cacp.dataset import AVAILABLE_N_FOLDS


class Dataset{}(ClassificationDatasetMinimalBase):  # do not change class declaration

    def folds(
        self,
        n_folds: AVAILABLE_N_FOLDS = 10,
        dob_scv: bool = True,
        categorical_to_numerical=True
    ) -> typing.Iterable[ClassificationFoldData]:  # do not change method declaration

        # change code below

        classes = 2
        features = 5
        instances = 100

        labels = np.array([label for label in range(classes)])
        x = np.random.rand(instances, features)
        y = np.random.choice(classes, instances)
        kf = KFold(n_splits=n_folds)

        folds = []

        for i, (train_index, test_index) in enumerate(kf.split(x), start=1):
            x_train, x_test = x[train_index], x[test_index]
            y_train, y_test = y[train_index], y[test_index]

            folds.append(ClassificationFoldData(
                index=i,
                labels=labels,
                x_test=x_test,
                y_test=y_test,
                x_train=x_train,
                y_train=y_train
            ))

        return iter(folds)
"""

KEEL_DATASET_CODE_TEMPLATE = """from pathlib import Path

from cacp import LocalClassificationDataset


class Dataset{}(LocalClassificationDataset):

    def __init__(self):
        super().__init__("{}", Path(r"{}"))
"""

CSV_DATASET_CODE_TEMPLATE = """from pathlib import Path

from cacp import LocalCsvClassificationDataset


class Dataset{}(LocalCsvClassificationDataset):

    def __init__(self):
        super().__init__("{}", Path(r"{}"))
"""


def _convert_from_document_to_custom_dataset(custom_dataset: Document) -> CustomDataset:
    if custom_dataset:
        custom_dataset["id"] = custom_dataset.doc_id
        custom_dataset["created at"] = datetime.fromtimestamp(custom_dataset["created_at"]).strftime(
            "%Y-%m-%d %H:%M:%S")
        custom_dataset["json_schema"] = {
            "title": custom_dataset["name"],
            "type": "object",
            "properties": {
            }
        }
    return custom_dataset


def _locate_id(custom_dataset_id: int, original_custom_dataset_id: typing.Optional[int] = None):
    if original_custom_dataset_id is None:
        original_custom_dataset_id = custom_dataset_id

    return f"cacp.gui.custom.datasets.dataset{custom_dataset_id}.Dataset{original_custom_dataset_id}"


def _code_path(custom_dataset_id: int) -> Path:
    return CUSTOM_DATASETS_CODE_DIR.joinpath(f"dataset{custom_dataset_id}.py")


def _save_code(custom_dataset_id: int, code: str):
    with _code_path(custom_dataset_id).open("w") as f:
        f.write(code)


def _remove_code(custom_dataset_id: int):
    _code_path(custom_dataset_id).unlink(missing_ok=True)


def get_all_custom_datasets() -> List[CustomDataset]:
    return [_convert_from_document_to_custom_dataset(e) for e in CUSTOM_DATASETS_DB.all() if e is not None]


def get_custom_dataset(custom_dataset_id: int) -> CustomDataset:
    return _convert_from_document_to_custom_dataset(CUSTOM_DATASETS_DB.get(doc_id=custom_dataset_id))


def add_custom_dataset() -> int:
    preview_prevent_modifications()
    new_custom_dataset: CustomDataset = dict()
    all_datasets = CUSTOM_DATASETS_DB.all()
    new_custom_dataset_id = 1 if len(all_datasets) == 0 else all_datasets[-1].doc_id + 1
    new_custom_dataset["name"] = f"Custom Dataset {new_custom_dataset_id}"
    new_custom_dataset["type"] = CustomDatasetType.CUSTOM_CODE
    new_custom_dataset["code"] = CUSTOM_DATASET_CODE_TEMPLATE.format(new_custom_dataset_id)
    new_custom_dataset["locate_id"] = _locate_id(new_custom_dataset_id)
    new_custom_dataset["created_at"] = datetime.now().timestamp()
    _save_code(new_custom_dataset_id, new_custom_dataset["code"])
    return CUSTOM_DATASETS_DB.insert(new_custom_dataset)


def update_custom_dataset(
    custom_dataset_id: int,
    name_value: str,
    type_value: CustomDatasetType,
    code_value: str,
    path_value: typing.Optional[str] = None
):
    preview_prevent_modifications()
    CUSTOM_DATASETS_DB.update({"name": name_value, "type": type_value, "code": code_value,
                               "path": path_value, "locate_id": _locate_id(custom_dataset_id)},
                              doc_ids=[custom_dataset_id])
    _save_code(custom_dataset_id, code_value)


def test_custom_dataset_code(custom_dataset_id: int, name_value: str, code_value: str):
    preview_prevent_modifications()
    error = None
    _save_code(0, code_value)
    try:
        ds = parse_dataset(
            {"name": name_value, "locate_id": _locate_id(0, custom_dataset_id), "code": code_value}
        )
        ds_fold = next(ds.folds())

        #  count classes
        labels = set()
        for _, y in ds:
            labels.add(y)
        number_of_classes = len(labels)

        # test if works for BATCH
        process_comparison_single(lambda n_inputs, n_classes: DummyClassifier(), "test", ds, ds_fold, DEFAULT_METRICS)

        # test if works for INCREMENTAL
        with tempfile.TemporaryDirectory() as tmpdirname:
            process_incremental_comparison_single(
                lambda n_inputs, n_classes: NoChangeClassifier(), "test",
                ds, number_of_classes, Path(tmpdirname), DEFAULT_INCREMENTAL_METRICS
            )
    except:
        error = traceback.format_exc()

    _remove_code(0)
    return error


def delete_custom_dataset(custom_dataset_id: int):
    preview_prevent_modifications()
    CUSTOM_DATASETS_DB.remove(doc_ids=[custom_dataset_id])
    _remove_code(custom_dataset_id)



import typing
import warnings
from pathlib import Path

import numpy as np
import pandas as pd
from scipy.stats import wilcoxon

from cacp.comparison import DEFAULT_METRICS
from cacp.util import to_latex


def bold_large_p_value(data: float, format_string="%.4f") -> str:
    """
    Makes large p-value in Latex table bold

    :param data: value
    :param format_string:
    :return: bolded values string

    """
    if data > 0.05:
        return "\\textbf{%s}" % format_string % data

    return "%s" % format_string % data


def process_wilcoxon_for_metric(current_algorithm: str, metric: str, result_dir: Path) -> pd.DataFrame:
    """
    Calculates the Wilcoxon signed-rank test for comparison results single metric.

    :param current_algorithm: current algorithm
    :param metric: comparison metric {auc, accuracy, precision, recall, f1}
    :param result_dir: results directory
    :return: DateFrame with wilcoxon values for metric

    """
    wilcoxon_dir = result_dir.joinpath('wilcoxon')
    wilcoxon_dir.mkdir(exist_ok=True, parents=True)
    metric_dir = wilcoxon_dir.joinpath(metric.lower())
    metric_dir.mkdir(exist_ok=True, parents=True)

    records = []
    df = pd.read_csv(result_dir.joinpath('comparison.csv'))
    algorithms = list(df['Algorithm'].unique())
    algorithms.remove(current_algorithm)
    current_alg_df = df[df['Algorithm'] == current_algorithm]
    for algorithm in algorithms:
        a_df = df[df['Algorithm'] == algorithm]
        alg1_values = current_alg_df[metric].values
        alg2_values = a_df[metric].values
        diff = alg1_values - alg2_values
        if np.all(diff == 0):
            w, p = 'invalid-data', 1
        else:
            w, p = wilcoxon(alg1_values, alg2_values)
        row = {
            current_algorithm: current_algorithm,
            'Algorithm': algorithm,
            'p-value': p,
        }
        records.append(row)

    df_r = pd.DataFrame(records)
    df_r.reset_index(drop=True, inplace=True)
    df_r.index += 1
    df_r.to_csv(metric_dir.joinpath(f'comparison_{current_algorithm}_result.csv'), index=True)
    f = metric_dir.joinpath(f'comparison_{current_algorithm}_result.tex').open('w')
    f.write(
        to_latex(df_r,
                 caption=f"Comparison of classifiers and {current_algorithm} "
                         f"using Wilcoxon signed-rank test for {metric}",
                 label=f'tab:{current_algorithm}_wilcoxon_{metric}_comparison',
                 )
    )
    return df_r


def process_wilcoxon(classifiers: typing.List[typing.Tuple[str, typing.Callable]], result_dir: Path,
                     metrics: typing.Sequence[typing.Tuple[str, typing.Callable]] = DEFAULT_METRICS):
    """
    Calculates the Wilcoxon signed-rank test for comparison results.

    :param classifiers: classifiers collection
    :param result_dir: results directory
    :param metrics: metrics collection

    """
    for current_algorithm, _ in classifiers:
        with warnings.catch_warnings():
            warnings.simplefilter(action='ignore', category=UserWarning)
            r_df = None
            for metric, _ in metrics:
                metric_wilcoxon = process_wilcoxon_for_metric(current_algorithm, metric, result_dir)
                if metric_wilcoxon.empty:
                    continue
                metric_wilcoxon = metric_wilcoxon.sort_values(by=['Algorithm'])
                if r_df is None:
                    r_df = metric_wilcoxon[['Algorithm']].copy()

                for c in metric_wilcoxon.columns[2:]:
                    r_df[f'{metric} {c}'] = metric_wilcoxon[c].values

            wilcoxon_dir = result_dir.joinpath('wilcoxon')
            wilcoxon_dir.mkdir(exist_ok=True, parents=True)

            if r_df is None:
                return

            r_df.reset_index(drop=True, inplace=True)
            r_df.index += 1
            r_df.to_csv(wilcoxon_dir.joinpath(f'comparison_{current_algorithm}.csv'), index=True)

            for metric, _ in metrics:
                col = f'{metric} p-value'
                r_df[col] = r_df[col].apply(lambda data: bold_large_p_value(data))

            wilcoxon_dir.joinpath(f'comparison_{current_algorithm}.tex').open('w').write(
                to_latex(r_df,
                         caption=f"Comparison of classifiers and {current_algorithm} using Wilcoxon signed-rank test",
                         label='tab:wilcoxon_comparison',
                         ))



from pydoc import locate
from typing import Dict, Callable, cast

from cacp.gui.external.river_library.metric import RiverMetricModel
from cacp.gui.external.sklearn_library.metric import SklearnMetricModel


def map_metric(metrics: list):
    return [
        {"#": i + 1, "docs_url": f"[{c.docs_url}]({c.docs_url})", "json_schema": c.json_schema, **c.dict()} for
        i, c in
        enumerate(sorted(metrics, key=lambda c: c.name))
    ]


def parse_metric(metric_dict: Dict):
    if "code" in metric_dict:  # custom metric
        return cast(Callable, locate(metric_dict["locate_id"], True))
    else:
        return cast(Callable, locate(metric_dict["id"]))


SKLEARN_METRICS = map_metric(SklearnMetricModel.all())
RIVER_METRICS = map_metric(RiverMetricModel.all())



import dash_ag_grid as dag
from dash import html, Output, Input, no_update, callback, State
from dash.dcc import Store

from cacp.gui.components.shared.utils import global_location_href_output
from cacp.gui.db.custom_datasets import get_all_custom_datasets, delete_custom_dataset
from cacp.gui.preview import preview_button_kwargs


class CustomDatasetsTable(html.Div):
    class ids:
        table = lambda aio_id: f"CustomDatasetsTable-table-{aio_id}"
        store = lambda aio_id: f"CustomDatasetsTable-store-{aio_id}"

    ids = ids

    @property
    def data(self):
        return get_all_custom_datasets()

    def __init__(
        self,
        aio_id,
        store_id=None
    ):
        self.table_id = self.ids.table(aio_id)
        self.store_id = store_id
        if store_id is None:
            self.store_id = self.ids.store(aio_id)

        super().__init__([
            dag.AgGrid(
                id=self.table_id,
                rowData=self.data,
                columnDefs=[
                    {"field": "id", "headerName": "#", "maxWidth": 100, "checkboxSelection": bool(store_id)},
                    {"field": "name", "maxWidth": None},
                    {"field": "type"},
                    {"field": "created at", "maxWidth": 200},
                    {
                        "field": "remove", "headerName": "", "maxWidth": 120, "sortable": False, "filter": False,
                        "resizable": False,
                        "cellRenderer": "Button",
                        "cellRendererParams": {
                            "buttonName": "Delete",
                            "buttonClassName": "btn btn-sm btn-danger",
                            "iconClassName": "bi bi-trash delete-button",
                            **preview_button_kwargs()
                        },
                    },
                    {
                        "field": "navigate", "headerName": "", "maxWidth": 120, "sortable": False, "filter": False,
                        "resizable": False,
                        "cellRenderer": "Button",
                        "cellRendererParams": {
                            "buttonName": "Open",
                            "buttonClassName": "btn btn-sm btn-primary",
                            "iconClassName": "bi bi-forward-fill"
                        },
                    },
                ],
                defaultColDef={"sortable": True, "filter": True, "resizable": True},
                dashGridOptions={"rowSelection": "single"} if store_id else None,
                columnSize="responsiveSizeToFit",
            ),
            Store(id=self.store_id)
        ])

        if store_id:
            @callback(
                Output(store_id, "data"),
                Input(self.table_id, "selectedRows"),
            )
            def selection_change(selected):
                return selected

        @callback(
            global_location_href_output(),
            Output(self.table_id, "rowData"),
            Input(self.table_id, "cellRendererData"),
            State(self.table_id, "rowData"),
            config_prevent_initial_callbacks="initial_duplicate"  # this will refresh every page reload
        )
        def on_datasets_renderer_data(
            custom_datasets_renderer_data,
            custom_datasets
        ):
            if custom_datasets_renderer_data:
                idx = custom_datasets_renderer_data["rowIndex"]
                custom_dataset_id = custom_datasets[idx]["id"]

                if custom_datasets_renderer_data["colId"] == "navigate":
                    return f"/custom_dataset/{custom_dataset_id}", no_update

                if custom_datasets_renderer_data["colId"] == "remove":
                    delete_custom_dataset(custom_dataset_id)

            return no_update, get_all_custom_datasets()



import dash
from dash import html, Input, callback, no_update

from cacp.gui.components.datasets.custom_datasets_table import CustomDatasetsTable
from cacp.gui.components.datasets.keel_datasets_table import KeelDatasetsTable
from cacp.gui.components.datasets.river_datasets_table import RiverDatasetsTable
from cacp.gui.components.shared.utils import global_location_href_output
from cacp.gui.db.custom_datasets import add_custom_dataset
from cacp.gui.preview import preview_button_kwargs

dash.register_page(__name__)

ADD_CUSTOM_DATASET_BUTTON_ID = "add_custom_dataset_button"

layout = html.Div([
    html.Div(html.Button("Add custom dataset", className="btn btn-primary", id=ADD_CUSTOM_DATASET_BUTTON_ID,
                         **preview_button_kwargs()),
             className="d-flex justify-content-end mb-4"),
    html.H5("Custom datasets"),
    CustomDatasetsTable("cdt"),
    html.H5("Keel datasets (BATCH, INCREMENTAL)", className="mt-4"),
    KeelDatasetsTable("kdt"),
    html.H5("River datasets (INCREMENTAL)", className="mt-4"),
    RiverDatasetsTable("rdt"),
])


@callback(
    global_location_href_output(),
    Input(ADD_CUSTOM_DATASET_BUTTON_ID, "n_clicks"),
    prevent_initial_call=True,
)
def add_custom_dataset_button(n_clicks):
    if n_clicks:
        custom_dataset_id = add_custom_dataset()
        return f"/custom_dataset/{custom_dataset_id}"

    return no_update



from cacp.dataset import \
    ClassificationDatasetMinimalBase, \
    ClassificationDatasetBase, \
    ClassificationDataset, \
    LocalClassificationDataset, \
    LocalCsvClassificationDataset, \
    ClassificationFoldData, \
    all_datasets

from cacp.run import run_experiment, run_incremental_experiment

__all__ = [
    'ClassificationDatasetMinimalBase',
    'ClassificationDatasetBase',
    'ClassificationDataset',
    'LocalClassificationDataset',
    'LocalCsvClassificationDataset',
    'ClassificationFoldData',
    'all_datasets',
    'run_experiment',
    'run_incremental_experiment'
]



import shutil
from datetime import datetime
from enum import Enum
from pathlib import Path
from typing import TypedDict, List, Dict, Optional

from tinydb import TinyDB
from tinydb.table import Document

from cacp.gui.custom.experiments import CUSTOM_EXPERIMENTS_DIR
from cacp.gui.db import DB_PATH
from cacp.gui.preview import preview_prevent_modifications


class ExperimentType(str, Enum):
    BATCH = "BATCH"
    INCREMENTAL = "INCREMENTAL"


class ExperimentStatus(str, Enum):
    RUNNING = "RUNNING"
    FINISHED = "FINISHED"
    FAILED = "FAILED"


class Experiment(TypedDict):
    id: int
    path: Optional[str]
    name: str
    datasets: List[Dict]
    classifiers: List[Dict]
    metrics: List[Dict]
    type: ExperimentType
    status: ExperimentStatus
    created_at: float


EXPERIMENTS_DB = TinyDB(CUSTOM_EXPERIMENTS_DIR / "experiments.json")
EXPERIMENTS_PATH = (DB_PATH / "experiments").resolve()


def _path(experiment_id: int) -> Path:
    return CUSTOM_EXPERIMENTS_DIR.joinpath(str(experiment_id))


def _convert_from_document_to_experiment(experiment: Document) -> Experiment:
    experiment["id"] = experiment.doc_id
    experiment["number of datasets"] = len(experiment["datasets"])
    experiment["number of classifiers"] = len(experiment["classifiers"])
    experiment["created at"] = datetime.fromtimestamp(experiment["created_at"]).strftime("%Y-%m-%d %H:%M:%S")
    experiment["path"] = str(_path(experiment.doc_id))
    return experiment


def get_all_experiments() -> List[Experiment]:
    return [_convert_from_document_to_experiment(e) for e in EXPERIMENTS_DB.all()]


def get_experiment(experiment_id: int) -> Experiment:
    return _convert_from_document_to_experiment(EXPERIMENTS_DB.get(doc_id=experiment_id))


def add_experiment(
    name_value: str, type_value: ExperimentType, selected_datasets: List[Dict], selected_classifiers: List[Dict],
    selected_metrics: List[Dict]
) -> int:
    preview_prevent_modifications()
    new_experiment: Experiment = dict()
    new_experiment["name"] = name_value
    new_experiment["type"] = type_value
    new_experiment["datasets"] = selected_datasets
    new_experiment["classifiers"] = selected_classifiers
    new_experiment["metrics"] = selected_metrics
    new_experiment["status"] = ExperimentStatus.RUNNING
    new_experiment["created_at"] = datetime.now().timestamp()
    experiment_id = EXPERIMENTS_DB.insert(new_experiment)
    return experiment_id


def update_experiment_status(experiment_id: int, experiment_status: ExperimentStatus):
    preview_prevent_modifications()
    EXPERIMENTS_DB.update({"status": experiment_status}, doc_ids=[experiment_id])


def delete_experiment(experiment_id: int):
    preview_prevent_modifications()
    shutil.rmtree(_path(experiment_id), ignore_errors=True)
    EXPERIMENTS_DB.remove(doc_ids=[experiment_id])



import contextlib
from functools import lru_cache
from typing import Type, Dict

import sklearn
from sklearn.base import ClassifierMixin
from sklearn.utils import all_estimators

from cacp.gui.external.shared.model import ClassModel, T
from cacp.gui.external.shared.type import to_id


class SklearnClassifierModel(ClassModel):
    name: str
    docs_url: str

    @classmethod
    def base_class(cls):
        return ClassifierMixin

    @classmethod
    def from_class(cls, source_class: Type) -> "SklearnClassifierModel":
        _id = to_id(source_class)
        docs_version_split = str(sklearn.__version__).split(".")
        docs_version = f"{docs_version_split[0]}.{docs_version_split[1]}"
        docs_split = _id.split(".")
        docs_name = ".".join(s for s in docs_split if not s.startswith("_"))
        model = cls(
            id=_id,
            name=source_class.__name__,
            docs_url=f"https://scikit-learn.org/{docs_version}/modules/generated/{docs_name}.html"
        )
        with contextlib.suppress(Exception):
            model.test()
            return model

    @classmethod
    @lru_cache
    def all_dict(cls) -> Dict[str, T]:
        result = {}
        for name, c in all_estimators(type_filter="classifier"):
            if model := cls.from_class(c):
                result[name] = model
        return result



import tempfile
import traceback
import typing
from datetime import datetime
from enum import Enum
from pathlib import Path
from typing import TypedDict, List

from tinydb import TinyDB
from tinydb.table import Document

from cacp import ClassificationDataset
from cacp.comparison import process_comparison_single, DEFAULT_METRICS, process_incremental_comparison_single, \
    DEFAULT_INCREMENTAL_METRICS
from cacp.gui.custom.classifiers import CUSTOM_CLASSIFIERS_CODE_DIR
from cacp.gui.external.classifier import parse_classifier
from cacp.gui.preview import preview_prevent_modifications


class CustomClassifierType(str, Enum):
    BATCH = "BATCH"
    INCREMENTAL = "INCREMENTAL"


class CustomClassifier(TypedDict):
    id: int
    locate_id: str
    name: str
    type: CustomClassifierType
    code: str
    created_at: float


CUSTOM_CLASSIFIERS_DB = TinyDB(CUSTOM_CLASSIFIERS_CODE_DIR / "classifiers.json")
CUSTOM_CLASSIFIER_BATCH_CODE_TEMPLATE = """import numpy as np
from sklearn.base import BaseEstimator, ClassifierMixin


#  Example BATCH classifier, returns always first class
class Classifier{}(BaseEstimator, ClassifierMixin):  # do not change class declaration

    def __init__(self):
        self.class_value = None

    def fit(
            self,
            X: np.ndarray,  # 2d array
            y: np.ndarray  # 1d array
    ):
        self.class_value = y[0]
        return self

    def predict(
            self,
            X: np.ndarray  # 1d array
    ):
        if self.class_value is None:
            raise Exception("Classifier not fitted")

        return np.full(len(X), self.class_value)
"""

CUSTOM_INCREMENTAL_CLASSIFIER_CODE_TEMPLATE = """import typing

from river import base


#  Example INCREMENTAL classifier, returns last seen class
class Classifier{0}(base.Classifier):  # do not change class declaration

    def __init__(self):
        self.last_class = None
        self.classes = set()

    def learn_one(self, x: typing.Dict, y: base.typing.ClfTarget, **kwargs) -> "Classifier{0}":
        self.last_class = y
        self.classes.add(y)
        return self

    def predict_one(self, x: typing.Dict) -> base.typing.ClfTarget:
        return self.last_class

    def predict_proba_one(self, x: typing.Dict) -> typing.Dict[base.typing.ClfTarget, float]:
        probabilities = dict()
        for c in self.classes:
            probabilities[c] = 0
        probabilities[self.last_class] = 1
        return probabilities
"""


def _convert_from_document_to_custom_classifier(custom_classifier: Document) -> CustomClassifier:
    if custom_classifier:
        custom_classifier["id"] = custom_classifier.doc_id
        custom_classifier["created at"] = datetime.fromtimestamp(custom_classifier["created_at"]).strftime(
            "%Y-%m-%d %H:%M:%S")
        custom_classifier["json_schema"] = {
            "title": custom_classifier["name"],
            "type": "object",
            "properties": {
            }
        }
    return custom_classifier


def _locate_id(custom_classifier_id: int, original_custom_classifier_id: typing.Optional[int] = None):
    if original_custom_classifier_id is None:
        original_custom_classifier_id = custom_classifier_id
    return f"cacp.gui.custom.classifiers.classifier{custom_classifier_id}.Classifier{original_custom_classifier_id}"


def _code_path(custom_classifier_id: int) -> Path:
    return CUSTOM_CLASSIFIERS_CODE_DIR.joinpath(f"classifier{custom_classifier_id}.py")


def _save_code(custom_classifier_id: int, code: str):
    with _code_path(custom_classifier_id).open("w") as f:
        f.write(code)


def _remove_code(custom_classifier_id: int):
    _code_path(custom_classifier_id).unlink(missing_ok=True)


def get_all_custom_classifiers() -> List[CustomClassifier]:
    return [_convert_from_document_to_custom_classifier(e) for e in CUSTOM_CLASSIFIERS_DB.all() if e is not None]


def get_custom_classifier(custom_classifier_id: int) -> CustomClassifier:
    return _convert_from_document_to_custom_classifier(CUSTOM_CLASSIFIERS_DB.get(doc_id=custom_classifier_id))


def add_custom_classifier() -> int:
    preview_prevent_modifications()
    new_custom_classifier: CustomClassifier = dict()
    all_classifiers = CUSTOM_CLASSIFIERS_DB.all()
    new_custom_classifier_id = 1 if len(all_classifiers) == 0 else all_classifiers[-1].doc_id + 1
    new_custom_classifier["name"] = f"Custom Classifier {new_custom_classifier_id}"
    new_custom_classifier["type"] = CustomClassifierType.BATCH
    new_custom_classifier["code"] = CUSTOM_CLASSIFIER_BATCH_CODE_TEMPLATE.format(new_custom_classifier_id)
    new_custom_classifier["locate_id"] = _locate_id(new_custom_classifier_id)
    new_custom_classifier["created_at"] = datetime.now().timestamp()
    _save_code(new_custom_classifier_id, new_custom_classifier["code"])
    return CUSTOM_CLASSIFIERS_DB.insert(new_custom_classifier)


def update_custom_classifier(
    custom_classifier_id: int,
    name_value: str,
    type_value: CustomClassifierType,
    code_value: str
):
    preview_prevent_modifications()
    CUSTOM_CLASSIFIERS_DB.update({"name": name_value, "type": type_value, "code": code_value},
                                 doc_ids=[custom_classifier_id])
    _save_code(custom_classifier_id, code_value)


def test_custom_classifier_code(custom_classifier_id: int, code_value: str, type_value: CustomClassifierType):
    preview_prevent_modifications()
    error = None
    _save_code(0, code_value)
    classifier_factory = parse_classifier({"locate_id": _locate_id(0, custom_classifier_id), "code": code_value})

    try:
        ds = ClassificationDataset("iris")
        ds_fold = next(ds.folds())
        if type_value == CustomClassifierType.BATCH:
            process_comparison_single(classifier_factory, "test", ds, ds_fold, DEFAULT_METRICS)
        elif type_value == CustomClassifierType.INCREMENTAL:
            with tempfile.TemporaryDirectory() as tmpdirname:
                process_incremental_comparison_single(
                    classifier_factory, "test",
                    ds, 3, Path(tmpdirname),
                    DEFAULT_INCREMENTAL_METRICS
                )
    except:
        error = traceback.format_exc()

    _remove_code(0)
    return error


def delete_custom_classifier(custom_classifier_id: int):
    preview_prevent_modifications()
    CUSTOM_CLASSIFIERS_DB.remove(doc_ids=[custom_classifier_id])
    _remove_code(custom_classifier_id)



import typing
from collections import defaultdict
from dataclasses import dataclass
from pathlib import Path

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd

from cacp.comparison import DEFAULT_METRICS, DEFAULT_INCREMENTAL_METRICS


def process_comparison_results_plots(result_dir: Path,
                                     metrics: typing.Sequence[typing.Tuple[str, typing.Callable]] = DEFAULT_METRICS):
    """
    Generates plots from comparison results.

    :param result_dir: results directory
    :param metrics: metrics collection

    """
    df_results = pd.read_csv(result_dir.joinpath('comparison.csv'))
    plot_dir = result_dir.joinpath('plot')
    plot_dir.mkdir(exist_ok=True, parents=True)

    def boxplot_sorted(df, by, column, file_suffix):
        df2 = pd.DataFrame({col: vals[column] for col, vals in df.groupby(by)})
        meds = df2.median().sort_values(ascending=False)
        df2[meds.index].boxplot(return_type="axes")
        plt.title('')
        plt.suptitle('')
        plt.ylabel(column)
        plt.xlabel('Algorithm')
        plt.xticks(rotation=90)
        plt.tight_layout()
        plt.savefig(plot_dir.joinpath(f'comparison_{column.lower()}{file_suffix}.eps'))
        plt.savefig(plot_dir.joinpath(f'comparison_{column.lower()}{file_suffix}.png'))
        plt.close()

    for metric, _ in metrics:
        boxplot_sorted(df_results, column=metric, by='Algorithm', file_suffix='_per_fold')
        boxplot_sorted(df_results.groupby(['Algorithm', 'Dataset']).mean().reset_index(level=0),
                       column=metric, by='Algorithm', file_suffix='_per_dataset')


@dataclass
class Line:
    x: np.ndarray
    y: np.ndarray
    label: str = ''


def process_comparison_results_incremental_plot(file_name: str, y_label: str, lines: typing.List[Line], plot_dir: Path):
    for line in lines:
        plt.plot(line.x, line.y, label=line.label)
    plt.xlabel('Number of samples')
    plt.ylabel(y_label)
    ax = plt.gca()
    if len(lines) > 1:
        ax.legend()
    ax.set_ylim([-0.05, 1.05])
    plt.savefig(plot_dir.joinpath(f'{file_name}.eps'))
    plt.savefig(plot_dir.joinpath(f'{file_name}.png'))
    plt.close()


def process_comparison_results_single_incremental_plot(classifier_name: str, dataset_name: str, metric: str,
                                                       df: pd.DataFrame, incremental_plot_dir: Path):
    """
    Generates plots from single incremental comparison results.

    :param classifier_name: classifier name
    :param dataset_name: dataset name
    :param metric: metric name
    :param df: result dataframe
    :param incremental_plot_dir: output plot directory

    """

    incremental_classifier_single_plot_dir = incremental_plot_dir.joinpath('single').joinpath(classifier_name)
    incremental_classifier_single_plot_dir.mkdir(exist_ok=True, parents=True)
    line = Line(df['index'], df[metric])
    process_comparison_results_incremental_plot(
        f"{classifier_name}_{dataset_name}_{metric.lower()}",
        metric,
        [line],
        incremental_classifier_single_plot_dir
    )


def process_comparison_results_incremental_plots(
    result_dir: Path,
    metrics: typing.Sequence[typing.Tuple[str, typing.Callable]] = DEFAULT_INCREMENTAL_METRICS
):
    """
    Generates plots from incremental comparison results.

    :param result_dir: results directory
    :param metrics: metrics collection

    """
    incremental_comparison_dir = result_dir.joinpath('incremental').joinpath('result')
    incremental_plot_dir = result_dir.joinpath('incremental').joinpath('plot')
    incremental_plot_dir.mkdir(exist_ok=True, parents=True)

    dataset_dfs = defaultdict(list)
    classifier_names = []
    metrics_names = [m for m, _ in metrics]

    for classifier_dir in incremental_comparison_dir.glob("*"):
        classifier_name = classifier_dir.stem
        classifier_names.append(classifier_name)
        for classifier_dataset_file in classifier_dir.glob("*.csv"):
            dataset_name = classifier_dataset_file.stem
            df = pd.read_csv(classifier_dataset_file)
            dataset_dfs[dataset_name].append(df)
            for metric in metrics_names:
                process_comparison_results_single_incremental_plot(
                    classifier_name,
                    dataset_name,
                    metric,
                    df,
                    incremental_plot_dir
                )

    for dataset_name, dfs in dataset_dfs.items():
        for metric in metrics_names:
            lines = [Line(df['index'], df[metric], classifier_names[i]) for i, df in enumerate(dfs)]
            process_comparison_results_incremental_plot(
                f"{dataset_name}_{metric.lower()}",
                metric,
                lines,
                incremental_plot_dir
            )



import dash_ag_grid as dag
from dash import html, callback, Output, Input
from dash.dcc import Store


class SelectedMetricsTable(html.Div):
    class ids:
        table = lambda aio_id: f"SelectedMetricsTable-table-{aio_id}"

    ids = ids

    def __init__(
        self,
        store_id,
        aio_id,
        init_data=None
    ):
        self.aio_id = aio_id

        if init_data is None:
            init_data = []

        super().__init__([
            dag.AgGrid(
                id=self.ids.table(self.aio_id),
                rowData=init_data,
                columnDefs=[
                    {"field": "name", "headerName": "Name"},
                    {"field": "docs_url", "cellRenderer": "markdown", "headerName": "Docs"},
                    {"field": "id", "cellRenderer": "markdown", "headerName": "Python path"},
                    {
                        "field": "name", "headerName": "", "maxWidth": 160, "sortable": False, "filter": False,
                        "resizable": False,
                        "cellRenderer": "Button",
                        "cellRendererParams": {
                            "buttonName": "Delete",
                            "buttonClassName": "btn btn-sm btn-danger",
                            "iconClassName": "bi bi-trash delete-button"
                        },
                    },
                ],
                defaultColDef={"sortable": True, "filter": True, "resizable": True},
                columnSize="responsiveSizeToFit",
            ),
            Store(id=store_id, data=init_data)
        ])

        @callback(
            Output(self.ids.table(self.aio_id), "rowData"),
            Input(store_id, "data"),

        )
        def selection_change(selected):
            return selected



import inspect
import os
import typing
from pathlib import Path
from timeit import default_timer as timer

import pandas as pd
import river
from joblib import delayed, Parallel
from river import metrics as river_metrics, stream, utils
from river.datasets.base import Dataset
from tqdm import tqdm

from cacp.dataset import ClassificationDatasetBase, ClassificationFoldData, AVAILABLE_N_FOLDS, \
    ClassificationFoldDataModifierBase, ClassificationFoldDataNormalizer
from cacp.util import accuracy, precision, recall, auc, f1

DEFAULT_METRICS = (('AUC', auc), ('Accuracy', accuracy), ('Precision', precision), ('Recall', recall), ('F1', f1))

DEFAULT_INCREMENTAL_METRICS = (
    ('AUC', river_metrics.ROCAUC), ('Accuracy', river_metrics.Accuracy), ('Precision', river_metrics.Precision),
    ('Recall', river_metrics.Recall), ('F1', river_metrics.F1)
)


def process_comparison_single(
    classifier_factory, classifier_name,
    dataset: ClassificationDatasetBase,
    fold: ClassificationFoldData,
    metrics: typing.Sequence[typing.Tuple[str, typing.Callable]],
) -> dict:
    """
    Runs comparison on single classifier and dataset.

    :param classifier_factory: classifier factory
    :param classifier_name: classifier name
    :param dataset: single dataset
    :param fold: fold data
    :param metrics: metrics collection
    :return: dictionary of calculated metrics and metadata

    """
    model = classifier_factory(fold.x_train.shape[1], len(fold.labels))
    train_start_time = timer()
    labels = fold.labels
    pred = None
    try:
        if 'classes' in inspect.getfullargspec(model.fit).args:
            model.fit(fold.x_train, fold.y_train, classes=labels.tolist())
        else:
            model.fit(fold.x_train, fold.y_train)
        train_time = timer() - train_start_time
        pred_start_time = timer()
        pred = model.predict(fold.x_test)
        pred_time = timer() - pred_start_time
    except Exception as e:
        train_time = 9999
        pred_time = 9999
        print(f"Error while running {classifier_name}, metrics will be set to 0", e)

    result = {
        'Dataset': dataset.name,
        'Algorithm': classifier_name,
        'Number of classes': len(set(fold.y_train)),
        'Train size': len(fold.x_train),
        'Test size': len(fold.x_test),
        'CV index': fold.index,
        'Train time [s]': train_time,
        'Prediction time [s]': pred_time
    }

    for (metric, metric_fun) in metrics:
        try:
            result[metric] = metric_fun(fold.y_test, pred, labels)
        except Exception as e:
            result[metric] = 0.
            print(f"Error while calculating {metric} for {classifier_name}, value will be set to 0", e)

    return result


def process_comparison(
    datasets: typing.List[ClassificationDatasetBase],
    classifiers: typing.List[typing.Tuple[str, typing.Callable]],
    result_dir: Path,
    metrics: typing.Sequence[typing.Tuple[str, typing.Callable]] = DEFAULT_METRICS,
    n_folds: AVAILABLE_N_FOLDS = 10,
    custom_fold_modifiers: typing.List[ClassificationFoldDataModifierBase] = None,
    dob_scv: bool = True,
    categorical_to_numerical=True,
    normalized: bool = False,
    progress=lambda progress, total: None,
):
    """
    Runs comparison for provided datasets and classifiers.

    :param datasets: dataset collection
    :param classifiers: classifiers collection
    :param result_dir: results directory
    :param metrics: metrics collection
    :param n_folds: number of folds {5,10}
    :param custom_fold_modifiers: custom fold modifiers that can change fold data before usage
    :param dob_scv: if folds distribution optimally balanced stratified cross-validation (DOB-SCV) should be used
    :param categorical_to_numerical: if dataset categorical values should be converted to numerical
    :param normalized: if the data should be normalized in range [0..1]
    :param progress: function that can be used to monitor progress

    """
    count = 0
    records = []
    df = None

    fold_modifiers = []

    if custom_fold_modifiers:
        fold_modifiers.extend(custom_fold_modifiers)

    if normalized:
        fold_modifiers.append(ClassificationFoldDataNormalizer())

    with tqdm(total=len(datasets) * n_folds, desc='Processing comparison', unit='fold') as pbar:
        progress(pbar.n, pbar.total)
        for dataset_idx, dataset in enumerate(datasets):
            for fold in dataset.folds(n_folds=n_folds, dob_scv=dob_scv,
                                      categorical_to_numerical=categorical_to_numerical):

                modified_fold = fold
                for fold_modifier in fold_modifiers:
                    modified_fold = fold_modifier.modify(modified_fold)

                rows = Parallel(n_jobs=len(classifiers))(
                    delayed(process_comparison_single)(c, c_n, dataset, modified_fold, metrics) for c_n, c in
                    classifiers
                )
                records.extend(rows)
                pbar.update(1)
                progress(pbar.n, pbar.total)

            df = pd.DataFrame(records)
            df = df.sort_values(by=['Dataset', 'Algorithm', 'CV index'])
            count += 1
            df.to_csv(result_dir.joinpath(f'comparison_{count}.csv'), index=False)
            if count > 1:
                prev_file = result_dir.joinpath(f'comparison_{count - 1}.csv')
                if os.path.isfile(prev_file):
                    os.remove(prev_file)

    if df is not None:
        prev_file = result_dir.joinpath(f'comparison_{count}.csv')
        if os.path.isfile(prev_file):
            os.remove(prev_file)
        df.to_csv(result_dir.joinpath('comparison.csv'), index=False)


def process_incremental_comparison_single(classifier_factory, classifier_name,
                                          dataset: typing.Union[
                                              ClassificationDatasetBase, Dataset
                                          ], number_of_classes: int, incremental_comparison_dir: Path,
                                          metrics: typing.Sequence[
                                              typing.Tuple[str, typing.Callable]] = DEFAULT_INCREMENTAL_METRICS
                                          ) -> dict:
    """
    Runs comparison on single classifier and dataset.

    :param classifier_factory: classifier factory
    :param classifier_name: classifier name
    :param dataset: single dataset
    :param number_of_classes: number of classes
    :param incremental_comparison_dir: incremental single results directory
    :param metrics: metrics collection
    :return: dictionary of calculated metrics and metadata

    """

    incremental_comparison_classifier_dir = incremental_comparison_dir.joinpath(classifier_name)
    incremental_comparison_classifier_dir.mkdir(exist_ok=True, parents=True)

    train_time = 0
    pred_time = 0

    train_size = 0
    dataset_name = "-"

    metric = river_metrics.base.Metrics([])

    try:
        dataset_type = type(dataset)
        if issubclass(dataset_type, ClassificationDatasetBase):
            dataset_name = dataset.name
            train_size = dataset.instances
        elif issubclass(dataset_type, Dataset):
            dataset_name = dataset.__class__.__name__.lower()
            train_size = dataset.n_samples

        metric = river_metrics.base.Metrics([m() for _, m in metrics])
        model = classifier_factory(train_size, number_of_classes)

        # Determine if predict_one or predict_proba_one should be used in case of a classifier
        if utils.inspect.isclassifier(model) and not metric.requires_labels:
            pred_func = model.predict_proba_one
        else:
            pred_func = model.predict_one

        records = []
        y_pred = None

        for i, x, y, *kwargs in stream.simulate_qa(dataset, None, None, copy=True):

            kwargs = kwargs[0] if kwargs else {}

            if y is None:
                # no ground truth, just make a prediction
                pred_start_time = timer()
                # predict
                y_pred = pred_func(x=x, **kwargs)
                pred_time_diff = timer() - pred_start_time
                pred_time += pred_time_diff
            else:
                # there's a ground truth, model and metric can be updated

                # update the metrics
                if y_pred != {} and y_pred is not None:
                    metric.update(y_true=y, y_pred=y_pred)
                    y_pred = max(y_pred, key=y_pred.get) if type(y_pred) is dict else y_pred
                    record = {
                        'index': i,
                        'y_true': y,
                        'y_pred': y_pred
                    }
                    for metric_idx, (metric_name, _) in enumerate(metrics):
                        record[metric_name] = metric.data[metric_idx].get()
                    records.append(record)

                learn_one_start_time = timer()
                # learn
                model.learn_one(x=x, y=y, **kwargs)
                learn_time_diff = timer() - learn_one_start_time
                train_time += learn_time_diff

        df = pd.DataFrame(records)
        df.to_csv(incremental_comparison_classifier_dir.joinpath(f'{dataset_name}.csv'), index=False)

    except Exception as e:
        print(f"Error while running {classifier_name}, metrics will be set to 0", e)

    result = {
        'Dataset': dataset_name,
        'Algorithm': classifier_name,
        'Number of classes': number_of_classes,
        'Train size': train_size,
        'Test size': train_size,
        'Train time [s]': train_time,
        'Prediction time [s]': pred_time
    }

    for metric_idx, (metric_name, _) in enumerate(metrics):
        try:
            value = metric.data[metric_idx].get()
            result[metric_name] = float(value)
        except Exception as e:
            print(f"Error while calculating {metric} for {classifier_name}, value will be set to 0", e)
            result[metric_name] = 0.

    return result


def process_incremental_comparison(
    datasets: typing.List[typing.Union[ClassificationDatasetBase, river.datasets.base.Dataset]],
    classifiers: typing.List[typing.Tuple[str, typing.Callable]],
    result_dir: Path,
    metrics: typing.Sequence[typing.Tuple[str, typing.Callable]] = DEFAULT_INCREMENTAL_METRICS,
    progress=lambda progress, total: None,
):
    """
    Runs comparison for provided datasets and incremental classifiers.

    :param datasets: dataset collection
    :param classifiers: classifiers collection
    :param result_dir: results directory
    :param metrics: metrics collection
    :param progress: function that can be used to monitor progress

    """

    incremental_comparison_dir = result_dir.joinpath('incremental').joinpath('result')
    incremental_comparison_dir.mkdir(exist_ok=True, parents=True)
    count = 0
    records = []
    df = None

    with tqdm(total=len(datasets), desc='Processing comparison', unit='dataset') as pbar:
        progress(pbar.n, pbar.total)
        for dataset_idx, dataset in enumerate(datasets):

            # preload dataset to prevent race conditions on file savings and count classes
            labels = set()
            for x, y in dataset:
                labels.add(y)
            number_of_classes = len(labels)

            rows = Parallel(n_jobs=len(classifiers))(
                delayed(process_incremental_comparison_single)(c, c_n, dataset, number_of_classes,
                                                               incremental_comparison_dir, metrics) for
                c_n, c in
                classifiers
            )
            records.extend(rows)
            pbar.update(1)
            progress(pbar.n, pbar.total)

            df = pd.DataFrame(records)
            df = df.sort_values(by=['Dataset', 'Algorithm'])
            count += 1
            df.to_csv(result_dir.joinpath(f'comparison_{count}.csv'), index=False)
            if count > 1:
                prev_file = result_dir.joinpath(f'comparison_{count - 1}.csv')
                if os.path.isfile(prev_file):
                    os.remove(prev_file)

    if df is not None:
        prev_file = result_dir.joinpath(f'comparison_{count}.csv')
        if os.path.isfile(prev_file):
            os.remove(prev_file)
        df.to_csv(result_dir.joinpath('comparison.csv'), index=False)



import typing
from pathlib import Path

import pandas as pd

from cacp.comparison import DEFAULT_METRICS
from cacp.util import to_latex


def process_comparison_results(result_dir: Path,
                               metrics: typing.Sequence[typing.Tuple[str, typing.Callable]] = DEFAULT_METRICS):
    """
    Processes comparison results, computes mean values for all metrics.

    :param result_dir: results directory
    :param metrics: metrics collection
    """
    df = pd.read_csv(result_dir.joinpath('comparison.csv'))

    gb = ['Algorithm']
    dfg = df.groupby(gb)
    df = dfg.mean(numeric_only=True)
    dfg_std = dfg.std(numeric_only=True)

    metrics_names = [name for name, _ in metrics]

    df_csv = df.copy(deep=True)
    for metric in metrics_names:
        df_csv.insert(list(df_csv.columns).index(metric) + 1, f'{metric} +/-', dfg_std[metric])

    columns = []
    for metric in metrics_names:
        columns.append(metric)
        columns.append(f'{metric} +/-')

    df_csv = df_csv[columns]
    df_csv = df_csv.sort_values(by=metrics_names, ascending=False)

    df_csv.reset_index(inplace=True)
    df_csv.index += 1
    df_csv.to_csv(result_dir.joinpath('comparison_result.csv'))

    df_tex = df_csv.copy(deep=True)
    for metric in metrics_names:
        metric_pm = f'{metric} +/-'
        df_tex[metric] = df_tex[metric].apply(lambda x: "{:1.3f}".format(x))
        df_tex[metric_pm] = df_tex[metric_pm].apply(lambda x: "{:1.3f}".format(x))
        df_tex[metric] = df_tex[metric].astype(str) + r'$\pm$' + df_tex[metric_pm].astype(str)
        del df_tex[metric_pm]

    f = result_dir.joinpath('comparison_result.tex').open('w')
    tex = to_latex(
        df_tex,
        caption='Results of comparison',
        label='tab:comparison',
    )
    f.write(tex)



from pathlib import Path

import dash_ag_grid as dag
import pandas as pd
from dash import Output


def csv_to_grid(path: Path, skip_first_column=False):
    columns = []
    row_data = []

    if path.exists():
        df = pd.read_csv(path)
        columns = df.columns
        if skip_first_column:
            columns = columns[1:]
        row_data = df[columns].to_dict("records")

    return dag.AgGrid(
        rowData=row_data,
        columnDefs=[
            {"field": c} for c in columns
        ],
        defaultColDef={"sortable": True, "filter": True, "resizable": True},
        columnSize="responsiveSizeToFit",
    )


def global_location_href_output():
    return Output(GLOBAL_LOCATION_ID, "href", allow_duplicate=True)


GLOBAL_LOCATION_ID = "global_location"



import dash_ag_grid as dag
from dash import html, callback, Output, Input
from dash.dcc import Store

from cacp.gui.external.classifier import RIVER_CLASSIFIERS


class RiverClassifiersTable(html.Div):
    class ids:
        table = lambda aio_id: f"RiverClassifiersTable-table-{aio_id}"
        store = lambda aio_id: f"RiverClassifiersTable-store-{aio_id}"

    ids = ids

    @property
    def data(self):
        return RIVER_CLASSIFIERS

    def __init__(
        self,
        aio_id,
        store_id=None
    ):

        self.store_id = store_id
        if store_id is None:
            self.store_id = self.ids.store(aio_id)

        super().__init__([
            dag.AgGrid(
                id=self.ids.table(aio_id),
                rowData=self.data,
                columnDefs=[
                    {"field": "#", "headerName": "#", "maxWidth": 100, "checkboxSelection": bool(store_id)},
                    {"field": "name", "headerName": "Name"},
                    {"field": "docs_url", "cellRenderer": "markdown", "headerName": "Docs", "maxWidth": None},
                    {"field": "id", "cellRenderer": "markdown", "headerName": "Python path", "maxWidth": None},
                ],
                defaultColDef={"maxWidth": 160, "sortable": True, "filter": True, "resizable": True},
                dashGridOptions={"rowSelection": "single"} if store_id else None,
                columnSize="responsiveSizeToFit",
            ),
            Store(id=self.store_id)
        ])

        if store_id:
            @callback(
                Output(store_id, "data"),
                Input(self.ids.table(aio_id), "selectedRows"),
            )
            def selection_change(selected):
                return selected



from pathlib import Path

import pandas as pd

from cacp.util import to_latex


def process_times(result_dir: Path):
    """
    Processes comparison results times.

    :param result_dir: results directory

    """
    df = pd.read_csv(result_dir.joinpath('comparison.csv'))
    time_dir = result_dir.joinpath('time')
    time_dir.mkdir(exist_ok=True, parents=True)
    gb = ['Algorithm']
    dfg = df.groupby(gb)
    df = dfg.mean(numeric_only=True)
    columns = ['Train time [s]', 'Prediction time [s]']
    df_csv = df[columns]
    df_csv = df_csv.sort_values(by=columns, ascending=True)
    df_csv.to_csv(time_dir.joinpath('comparison.csv'))

    df_tex = df_csv.copy(deep=True)
    df_tex.reset_index(drop=False, inplace=True)
    df_tex.index += 1
    tex = to_latex(
        df_tex,
        caption='Results of time comparison',
        label='tab:time_comparison',
    )

    with time_dir.joinpath('comparison.tex').open('w') as f:
        f.write(tex)



import dash_ag_grid as dag
from dash import html, callback, Output, Input
from dash.dcc import Store

from cacp.gui.external.river_library.dataset import RiverDatasetModel


class RiverDatasetsTable(html.Div):
    class ids:
        table = lambda aio_id: f"RiverDatasetsTable-table-{aio_id}"
        store = lambda aio_id: f"RiverDatasetsTable-store-{aio_id}"

    ids = ids

    @property
    def data(self):
        valid_datasets = sorted([rd for rd in RiverDatasetModel.all() if rd.classes > 0 and rd.samples > 0],
                                key=lambda x: x.name)
        return [
            {"#": i + 1, **frd.dict(), "docs_url": f"[{frd.docs_url}]({frd.docs_url})",
             "json_schema": frd.json_schema} for i, frd
            in enumerate(valid_datasets)
        ]

    def __init__(
        self,
        aio_id,
        store_id=None
    ):
        self.store_id = store_id
        if store_id is None:
            self.store_id = self.ids.store(aio_id)

        super().__init__([
            dag.AgGrid(
                id=self.ids.table(aio_id),
                rowData=self.data,
                columnDefs=[
                    {"field": "#", "headerName": "#", "maxWidth": 100, "checkboxSelection": bool(store_id)},
                    {"field": "name", "headerName": "Name"},
                    {"field": "docs_url", "cellRenderer": "markdown", "headerName": "Docs", "maxWidth": None},
                    {"field": "samples", "headerName": "Instances", "filter": "agNumberColumnFilter"},
                    {"field": "features", "headerName": "Features", "filter": "agNumberColumnFilter"},
                    {"field": "classes", "headerName": "Classes", "filter": "agNumberColumnFilter"},
                ],
                defaultColDef={"maxWidth": 160, "sortable": True, "filter": True, "resizable": True},
                dashGridOptions={"rowSelection": "single"} if store_id else None,
                columnSize="responsiveSizeToFit",
            ),
            Store(id=self.store_id)
        ])

        if store_id:
            @callback(
                Output(store_id, "data"),
                Input(self.ids.table(aio_id), "selectedRows"),
            )
            def selection_change(selected):
                return selected



from pathlib import Path

import dash
import dash_bootstrap_components as dbc
from PIL import Image
from dash import html, Output, Input, no_update, callback, dcc
from dash.dcc import Store

from cacp.gui.components.experiments.details.processor import ExperimentProcessor
from cacp.gui.components.shared.utils import csv_to_grid
from cacp.gui.db.experiments import get_experiment, Experiment, ExperimentStatus, ExperimentType
from cacp.gui.preview import is_preview

dash.register_page(__name__, path_template="/experiment/<experiment_id>")

EXPERIMENT_STORE_ID = "experiment_store"
EXPERIMENT_RESULTS_ID = "experiment_results"
EXPERIMENT_RESULTS_PATH_ID = "experiment_results_path"


def show_experiment(experiment: Experiment):
    experiment_path = Path(experiment["path"])

    incremental_components = html.Div([])
    if experiment["type"] == ExperimentType.INCREMENTAL:
        incremental_components = html.Div([
            html.H5("Incremental plots", className="mt-4"),
            html.Div([
                html.Div([
                    html.P(plot.name.split("_")[0].upper() + " dataset", className="image-title"),
                    html.Div(html.Img(src=Image.open(plot)), className="d-flex justify-content-center"),
                ], className="col mt-5 position-relative") for plot in
                experiment_path.joinpath("incremental").joinpath("plot").glob("*.png")
            ], className="row justify-content-md-center")
        ])

    return html.Div([
        html.H5("Directory", className="mt-4"),
        dbc.InputGroup([
            dbc.Input(
                id=EXPERIMENT_RESULTS_PATH_ID,
                value=experiment["path"],
                readonly=True
            ),
            dcc.Clipboard(
                target_id=EXPERIMENT_RESULTS_PATH_ID,
                title="copy",
                style={
                    "margin-left": 10,
                    "fontSize": 25,
                },
            ),
        ]),
        html.H5("Datasets", className="mt-4"),
        csv_to_grid(experiment_path / "info" / "datasets.csv", skip_first_column=True),
        html.H5("Classifiers", className="mt-4"),
        csv_to_grid(experiment_path / "info" / "classifiers.csv", skip_first_column=True),
        html.H5("Comparison results", className="mt-4"),
        csv_to_grid(experiment_path / "comparison_result.csv", skip_first_column=True),
        html.H5("Comparison results winner", className="mt-4"),
        csv_to_grid(experiment_path / "winner" / "comparison.csv", skip_first_column=True),
        html.H5("Comparison times", className="mt-4"),
        csv_to_grid(experiment_path / "time" / "comparison.csv"),
        html.H5("Wilcoxon results", className="mt-4"),
        html.Div([
            html.Div([
                html.P(csv.name.split("_")[1].split(".")[0] + " vs:"),
                csv_to_grid(csv, True),
                html.Br(),
                html.Br()
            ]) for csv in experiment_path.joinpath("wilcoxon").glob("comparison_*.csv")
        ]),
        html.H5("Result plots", className="mt-4"),
        html.Div([
            html.Div([
                html.Img(src=Image.open(plot)),
                html.Br(),
                html.Br()
            ], className="col mt-5") for plot in experiment_path.joinpath("plot").glob("comparison_*_per_fold.png")
        ], className="row justify-content-md-center"),
        incremental_components
    ])


EXPERIMENT_PROCESSOR = ExperimentProcessor(EXPERIMENT_STORE_ID)


def layout(experiment_id=None):
    experiment = get_experiment(experiment_id) if experiment_id else None
    view = html.Div([
        EXPERIMENT_PROCESSOR if experiment and experiment[
            "status"] == ExperimentStatus.RUNNING and not is_preview() else "",
        html.Div(show_experiment(experiment) if experiment else [], id=EXPERIMENT_RESULTS_ID),
        Store(id=EXPERIMENT_STORE_ID, data=experiment),
    ])
    return view


@callback(
    Output(EXPERIMENT_RESULTS_ID, "children"),
    Input(EXPERIMENT_STORE_ID, "data")
)
def on_experiment_changed(experiment: Experiment):
    if experiment:
        return show_experiment(experiment)
    return no_update



import dash
from dash import html, callback, Input, no_update

from cacp.gui.components.metrics.custom_metrics_table import CustomMetricsTable
from cacp.gui.components.metrics.river_metrics_table import RiverMetricsTable
from cacp.gui.components.metrics.sklearn_metrics_table import SklearnMetricsTable
from cacp.gui.components.shared.utils import global_location_href_output
from cacp.gui.db.custom_metrics import add_custom_metric
from cacp.gui.preview import preview_button_kwargs

dash.register_page(__name__)

ADD_CUSTOM_CLASSIFIER_BUTTON_ID = "add_custom_metric_button"

layout = html.Div([
    html.Div(
        html.Button("Add custom metric", className="btn btn-primary", id=ADD_CUSTOM_CLASSIFIER_BUTTON_ID,
                    **preview_button_kwargs()),
        className="d-flex justify-content-end mb-4"),
    html.H5("Custom metrics"),
    CustomMetricsTable("cmt"),
    html.H5("Wrapped Sklearn metrics (BATCH)", className="mt-4"),
    html.P(
        "To ensure operation on as many data sets as possible the metrics have been wrapped. "
        "Default parameters if occurred were set to average='weighted' and zero_division=0."),
    SklearnMetricsTable("smt"),
    html.H5("River metrics (INCREMENTAL)", className="mt-4"),
    RiverMetricsTable("rmt"),
])


@callback(
    global_location_href_output(),
    Input(ADD_CUSTOM_CLASSIFIER_BUTTON_ID, "n_clicks"),
    prevent_initial_call=True,
)
def add_custom_metric_button(n_clicks):
    if n_clicks:
        custom_metric_id = add_custom_metric()
        return f"/custom_metric/{custom_metric_id}"
    return no_update



import dash_bootstrap_components as dbc
import dash_rjsf
from dash import html, Output, Input, callback, State, ctx, no_update
from dash.dcc import Store

SUBMIT_BUTTON_OPTIONS_NAME = "ui:submitButtonOptions"


class SelectionModal(html.Span):
    class ids:
        modal_open_button = lambda aio_id: f"SelectionModal-modal_open_button-{aio_id}"
        modal_close_button = lambda aio_id: f"SelectionModal-modal_close_button-{aio_id}"
        modal_add_all_button = lambda aio_id: f"SelectionModal-modal_add_all_button-{aio_id}"
        modal = lambda aio_id: f"SelectionModal-modal-{aio_id}"
        selection_store = lambda aio_id: f"SelectionModal-selection_store-{aio_id}"
        rjsf = lambda aio_id: f"SelectionModal-rjsf-{aio_id}"

    ids = ids

    def __init__(
        self,
        name,
        source_table,
        aio_id,
        button_kwargs=None,
    ):

        if button_kwargs is None:
            button_kwargs = {}

        self.aio_id = aio_id
        self.store_id = f"{aio_id}-store"

        table = source_table(store_id=self.ids.selection_store(aio_id), aio_id=aio_id)

        super().__init__([
            dbc.Button(name, id=self.ids.modal_open_button(aio_id), n_clicks=0, outline=True, color="primary",
                       **button_kwargs),

            dbc.Modal(
                [
                    dbc.ModalHeader(dbc.ModalTitle(name)),
                    dbc.ModalBody([
                        dbc.Button("Submit all using default properties", id=self.ids.modal_add_all_button(aio_id),
                                   outline=True,
                                   color="primary"),
                        html.Br(),
                        html.Br(),
                        table,
                        html.Br(),
                        dash_rjsf.DashRjsf(
                            id=self.ids.rjsf(aio_id),
                            schema={
                            },
                            uiSchema={SUBMIT_BUTTON_OPTIONS_NAME: {
                                "submitText": "Add",
                                "norender": False,
                                "props": {
                                    "disabled": False,
                                    "className": "rxjs-submit"
                                }
                            }}
                        )
                    ], className="selection-modal-body"),
                    Store(id=self.store_id)
                ],
                id=self.ids.modal(aio_id),
                size="xl",
                is_open=False,
            ),

        ], )

        @callback(
            Output(self.ids.rjsf(aio_id), "value"),
            Output(self.ids.rjsf(aio_id), "schema"),
            Output(self.ids.rjsf(aio_id), "uiSchema"),
            Input(self.ids.selection_store(aio_id), "data")
        )
        def on_table_selection(data):
            if data:
                schema = data[0]["json_schema"]
                if schema["properties"]:
                    schema["title"] = "Provide parameters for " + schema["title"]
                else:
                    schema["title"] = ""
                return None, schema, {SUBMIT_BUTTON_OPTIONS_NAME: {
                    "submitText": "Submit",
                    "norender": False,
                    "props": {
                        "disabled": False,
                        "className": "rxjs-submit"
                    }
                }}
            return None, {}, {SUBMIT_BUTTON_OPTIONS_NAME: {
                "submitText": "Submit",
                "norender": False,
                "props": {
                    "disabled": True,
                    "className": "rxjs-submit"
                }
            }}

        @callback(
            Output(self.store_id, "data"),
            Output(self.ids.modal(aio_id), "is_open"),
            Input(self.ids.rjsf(aio_id), "value"),
            State(self.ids.selection_store(aio_id), "data"),
            Input(self.ids.modal_open_button(aio_id), "n_clicks"),
            Input(self.ids.modal_add_all_button(aio_id), "n_clicks"),
            Input(self.ids.rjsf(aio_id), "n_clicks"),

        )
        def on_open_click_or_selection(values, selected, open_button_n_clicks, _add_all_button_n_clicks,
                                       _values_submit_n_clicks):
            new_data = no_update
            new_modal_is_open = no_update

            if ctx.triggered_id == self.ids.modal_open_button(aio_id) and open_button_n_clicks:
                new_modal_is_open = True

            # add single
            if selected and ctx.triggered_id == self.ids.rjsf(aio_id) and values is not None:
                new_data = [dict(init_values=values, **s) for s in selected]
                # TODO: validate if can be initialized with user params
                new_modal_is_open = False

            # submit all
            if ctx.triggered_id == self.ids.modal_add_all_button(aio_id):
                new_data = [dict(init_values={}, **d) for d in table.data]
                new_modal_is_open = False

            return new_data, new_modal_is_open



import logging

import click
import dash_bootstrap_components as dbc
import diskcache
import flask
from dash import Dash, html
from dash import DiskcacheManager
from dash import page_container
from dash.dcc import Location

from cacp.gui.components.shared.sidebar import sidebar_component
from cacp.gui.components.shared.utils import GLOBAL_LOCATION_ID
from cacp.gui.db import DB_PATH
from cacp.gui.preview import is_preview

CACHE_PATH = DB_PATH / ".cache"
CACHE_PATH.mkdir(exist_ok=True, parents=True)
BACKGROUND_CALLBACK_MANAGER = DiskcacheManager(diskcache.Cache(CACHE_PATH))


@click.command()
@click.option("--debug", "-d", is_flag=True, show_default=True, default=False, type=bool)
@click.option("--host", "-h", show_default=True, default="127.0.0.1", type=str)
@click.option("--port", "-p", show_default=True, default=8050, type=int)
def start(debug: bool = False, host="127.0.0.1", port=8050):
    log_level = logging.INFO if debug else logging.ERROR
    flask_log = logging.getLogger("werkzeug")
    flask_log.setLevel(log_level)
    flask.cli.show_server_banner = lambda *_args, **_kwargs: None

    app = Dash(
        __name__,
        external_stylesheets=[dbc.themes.COSMO, dbc.icons.BOOTSTRAP],
        use_pages=True,
        background_callback_manager=BACKGROUND_CALLBACK_MANAGER
    )
    app.layout = html.Div([
        Location(id=GLOBAL_LOCATION_ID),
        sidebar_component(),
        html.Div(id="page-content", style={
            "margin-left": "18rem",
            "margin-right": "2rem",
            "padding": "2rem 1rem",
        }, children=[page_container]),  # https://cacp.readthedocs.io/

        dbc.Alert(
            html.Div([
                html.Span(
                    "CACP is in PREVIEW mode, some actions are locked. "
                    "To enable them, run an instance on your machine "),
                html.A("https://cacp.readthedocs.io/", href="https://cacp.readthedocs.io/", target="_blank")
            ]),
            color="info",
            dismissable=True,
            is_open=True,
            style={
                "position": "fixed",
                "top": 20,
                "left": 0,
                "right": 0,
                "width": "50%",
                "margin": "0 auto"
            }
        ) if is_preview() else "",
    ])
    app.logger.setLevel(log_level)

    print(f"CACP stared on http://{host}:{port}/")
    app.run_server(debug=debug, host=host, port=port)


if __name__ == "__main__":
    start()



from functools import lru_cache
from typing import Dict, List

from cacp.gui.external.shared.model import T, BaseAppModel
from cacp.gui.external.shared.type import to_id
from cacp.util import matthews_corrcoef, auc, accuracy, precision, recall, f1

METRICS = (
    ("AUC", auc, "https://scikit-learn.org/stable/modules/generated/sklearn.metrics.roc_auc_score.html"),
    ("Accuracy", accuracy, "https://scikit-learn.org/stable/modules/generated/sklearn.metrics.accuracy_score.html"),
    ("Precision", precision, "https://scikit-learn.org/stable/modules/generated/sklearn.metrics.precision_score.html"),
    ("Recall", recall, "https://scikit-learn.org/stable/modules/generated/sklearn.metrics.recall_score.html"),
    ("F1", f1, "https://scikit-learn.org/stable/modules/generated/sklearn.metrics.f1_score.html"),
    ("Matthews correlation coefficient", matthews_corrcoef,
     "https://scikit-learn.org/stable/modules/generated/sklearn.metrics.matthews_corrcoef.html")
)


class SklearnMetricModel(BaseAppModel):
    id: str
    name: str
    docs_url: str

    @property
    def json_schema(self):
        return {"title": self.name, "type": "object", "properties": {}}

    @classmethod
    @lru_cache
    def all_dict(cls) -> Dict[str, T]:
        result = {}
        for name, function, docs_url in METRICS:
            model_id = to_id(function)
            result[model_id] = cls(
                id=model_id,
                name=name,
                docs_url=docs_url
            )

        return result

    @classmethod
    @lru_cache
    def all(cls) -> List[T]:
        return list(cls.all_dict().values())

    @classmethod
    @lru_cache
    def get_by_id(cls, _id: str) -> T:
        return cls.all_dict().get(_id)



from pydoc import locate
from typing import Dict, Callable, cast

from cacp import ClassificationDataset


def parse_dataset(dataset_dict: Dict):
    ds = None
    if "code" in dataset_dict:  # custom dataset
        dataset_type = cast(Callable, locate(dataset_dict["locate_id"], True))
        ds = dataset_type()
    elif dataset_dict["id"] == "cacp.dataset.ClassificationDataset":
        ds = ClassificationDataset(dataset_dict["Name"])
    else:
        dataset_type = cast(Callable, locate(dataset_dict["id"]))
        ds = dataset_type(**dataset_dict["init_values"])

    if not hasattr(ds, "name"):
        ds.name = dataset_dict.get("name", type(ds))
    return ds



from __future__ import annotations

import uuid
from abc import ABC, abstractmethod
from contextlib import suppress
from dataclasses import dataclass
from functools import lru_cache, cached_property
from inspect import signature, Parameter, isabstract
from pydoc import locate
from typing import (
    Deque, Dict, FrozenSet, List, Optional, Sequence, Set, Tuple, Union, get_args
)
from typing import Type, get_origin

from pydantic import BaseModel, BaseConfig
from pydantic.fields import ModelField
from river.base import Classifier, DriftDetector, Estimator
from river.datasets.base import Dataset
from river.metrics.base import Metric
from river.optim.base import Initializer, Scheduler, Optimizer

from cacp.gui.external.shared.type import get_all_non_abstract_subclasses, is_primitive_type, to_id

NONE_TYPE = type(None)
SEARCHABLE = [
    Dataset,
    Classifier,
    DriftDetector,
    Optimizer,
    Metric,
    Estimator,
    Initializer,
    Scheduler
]


@dataclass
class ToPydanticMapper(ABC):
    parameter: Parameter
    parent_type: Type

    @abstractmethod
    def can_map(self) -> bool:
        pass

    @abstractmethod
    def map(self) -> ModelField:
        pass

    @cached_property
    def annotation(self):
        args = get_args(self.parameter.annotation)
        if len(args) == 2:
            if args[0] is NONE_TYPE:  # handle type.Optional
                return args[1]
            elif args[1] is NONE_TYPE:  # handle type.Optional
                return args[0]
        return parse_type(self.parameter.annotation, self.parent_type)

    @cached_property
    def default(self):
        return self.parameter.default if is_primitive_type(
            self.annotation) and self.parameter.default != self.parameter.empty else None


@dataclass
class SearchableToPydanticMapper(ToPydanticMapper):
    searchable: Optional[Type] = None

    def can_map(self) -> bool:
        annotation = self.annotation
        with suppress(TypeError):
            for searchable in SEARCHABLE:
                if issubclass(annotation, searchable):
                    self.searchable = searchable
                    return True

        return False

    def map(self) -> ModelField:
        annotation = self._searchable_type_to_pseudo_pydantic_model(self.searchable, self.annotation)
        return ModelField.infer(
            name=self.parameter.name,
            value=None,
            annotation=annotation,
            class_validators=None,
            config=BaseConfig,
        )

    @staticmethod
    @lru_cache
    def _searchable_type_to_pseudo_pydantic_model(searchable, annotation):
        model = new_pseudo_pydantic_model(f"{searchable.__name__}__{uuid.uuid4().hex}")
        subclasses = get_all_non_abstract_subclasses(annotation)
        if not subclasses:
            subclasses.append(annotation)
        pseudo_subclasses = [
            new_pseudo_pydantic_model(f"{to_id(x)}__{uuid.uuid4().hex}") for x in
            subclasses
        ]
        p_annotation = Union[tuple(pseudo_subclasses)]
        model.__fields__["option"] = ModelField.infer(
            name="option",
            value=None,
            annotation=p_annotation,
            class_validators=None,
            config=BaseConfig,
        )
        return model


@dataclass
class IterableToPydanticMapper(ToPydanticMapper):
    ITERABLE_TYPES = [
        Deque,
        Dict,
        dict,
        FrozenSet,
        List,
        list,
        Optional,  # Union[x,None]
        Sequence,
        Set,
        set,
        Tuple,
        Union,
    ]

    def can_map(self) -> bool:
        annotation = self.annotation
        origin = get_origin(annotation)
        return origin in self.ITERABLE_TYPES or \
            annotation in self.ITERABLE_TYPES or \
            (type(annotation) is str and "|" in annotation)

    def map(self) -> ModelField:
        args = get_args(self.annotation)
        if len(args) == 1 and args[0] is NONE_TYPE:  # handle type.Optional
            annotation = type_to_pseudo_pydantic_model(args[1], self.parameter.name)
        elif len(args) == 2 and args[1] is NONE_TYPE:  # handle type.Optional
            annotation = type_to_pseudo_pydantic_model(args[0], self.parameter.name)
        elif len(args) == 1:
            annotation = self.annotation
        else:  # handle other types
            if type(self.annotation) is str and "|" in self.annotation:
                type_strings = [x.strip() for x in self.annotation.split("|")]
                type_strings_without_none = [x for x in type_strings if x != "None"]
                pseudo_values = [type_to_pseudo_pydantic_model(x, self.parameter.name) for x in
                                 type_strings_without_none]
            else:
                pseudo_values = [type_to_pseudo_pydantic_model(x, self.parameter.name) for x in args]
            annotation = Union[tuple(pseudo_values)]

        return ModelField.infer(
            name=self.parameter.name,
            value=self.default,
            annotation=annotation,
            class_validators=None,
            config=BaseConfig,
        )


@dataclass
class AnyAbstractToPydanticMapper(ToPydanticMapper):

    def can_map(self) -> bool:
        return isabstract(self.annotation)

    def map(self) -> ModelField:
        subclasses = get_all_non_abstract_subclasses(self.annotation)
        pseudo_subclasses = [type_to_pseudo_pydantic_model(x, self.parameter.name) for x in subclasses]
        return ModelField.infer(
            name=self.parameter.name,
            value=self.default,
            annotation=Union[tuple(pseudo_subclasses)],
            class_validators=None,
            config=BaseConfig,
        )


@dataclass
class AnyToPydanticMapper(ToPydanticMapper):

    def can_map(self) -> bool:
        if "typing.Callable" in str(self.annotation):
            return False
        return True

    def map(self) -> ModelField:
        annotation = type_to_pseudo_pydantic_model(self.annotation, self.parameter.name, self.parent_type)
        return ModelField.infer(
            name=self.parameter.name,
            value=self.default,
            annotation=annotation,
            class_validators=None,
            config=BaseConfig,
        )


@dataclass
class DocToPydanticMapper:

    @classmethod
    def map(cls, class_type, parameter) -> ModelField:
        doc_string = class_type.__doc__
        if doc_string:
            parameter_in_doc_string = f"{parameter.name} :"
            for line in doc_string.split("\n"):
                line = line.strip()
                if line.startswith(parameter_in_doc_string):
                    possible_primitive = line.replace(parameter_in_doc_string, "").strip().split(",")[0]
                    if possible_primitive in ["bool", "int", "float", "str"]:
                        return ModelField.infer(
                            name=parameter.name,
                            value=None,
                            annotation=locate(possible_primitive),
                            class_validators=None,
                            config=BaseConfig,
                        )


MAPPER_CLASSES = [
    SearchableToPydanticMapper,
    IterableToPydanticMapper,
    AnyAbstractToPydanticMapper,
    AnyToPydanticMapper
]


def new_pseudo_pydantic_model(name: str):
    class Model(BaseModel):
        pass

    Model.__name__ = name
    return Model


def parse_type(t: Union[Type, str], parent_type: Type = None):
    if type(t) is str:
        located_t = locate(t)
        if is_primitive_type(located_t):
            return located_t
        elif located_t:
            t = located_t
        # elif parent_type is not None and "." in t:
        #
        #     parts = t.split(".")
        #     located_t = locate(parent_type.__module__)
        #     for p in parts:
        #         ors = p.split("|")
        #         if len(ors) > 0:
        #             located_t = Union[tuple(located_t.__getattribute__(o) for o in ors)]
        #         else:
        #             located_t = located_t.__getattribute__(p)
        #     t = located_t
    return t


@lru_cache
def type_to_pseudo_pydantic_model(t: Union[Type, str], alternative_name: str = None, parent_type: Type = None):
    t = parse_type(t, parent_type)

    if is_primitive_type(t):
        return t

    if t in [None, "None"]:
        return type(None)

    model = new_pseudo_pydantic_model(t.__name__ if hasattr(t, "__name__") else alternative_name)
    sig = signature(t.__init__)
    init_parameters: List[Parameter] = list(sig.parameters.values())[1:]
    for parameter in init_parameters:
        print(parameter.annotation)

        # model.__fields__[parameter.name] = type_to_pseudo_pydantic_model(t)

        # if parameter.annotation == parameter.empty and parameter.default != parameter.empty:
        #     potential_annotation_type = type(parameter.default)
        #     if is_primitive_type(potential_annotation_type):
        #         parameter._annotation = potential_annotation_type
        # if parameter.annotation != parameter.empty:
        #     for mapper_class in MAPPER_CLASSES:
        #         mapper = mapper_class(parameter, t)
        #         if mapper.can_map():
        #             field = mapper.map()
        #             if field:
        #                 model.__fields__[parameter.name] = field
        #             break
        # else:
        #     field = DocToPydanticMapper.map(t, parameter)
        #     if field:
        #         model.__fields__[parameter.name] = field

    return model
